#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test for demo purposes."""
import os
import shutil
import subprocess
import tempfile
from collections import defaultdict

import numpy
import pytest
import yaml

import orion.core.cli
import orion.core.io.experiment_builder as experiment_builder
from orion.core.cli.hunt import workon
from orion.core.io.database.ephemeraldb import EphemeralDB
from orion.storage.base import get_storage
from orion.storage.legacy import Legacy
from orion.testing import OrionState


def test_demo_with_default_algo_cli_config_only(storage, monkeypatch):
    """Check that random algorithm is used, when no algo is chosen explicitly."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    orion.core.cli.main(
        [
            "hunt",
            "-n",
            "default_algo",
            "--max-trials",
            "5",
            "./black_box.py",
            "-x~uniform(-50, 50)",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "default_algo"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    assert exp["name"] == "default_algo"
    assert exp["pool_size"] == 1
    assert exp["max_trials"] == 5
    assert exp["max_broken"] == 3
    assert exp["algorithms"] == {"random": {"seed": None}}
    assert "user" in exp["metadata"]
    assert "datetime" in exp["metadata"]
    assert "orion_version" in exp["metadata"]
    assert "user_script" in exp["metadata"]
    assert exp["metadata"]["user_args"] == ["./black_box.py", "-x~uniform(-50, 50)"]

    trials = list(storage.fetch_trials(uid=exp["_id"]))
    assert len(trials) <= 10
    assert trials[-1].status == "completed"


def test_demo(storage, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "./black_box.py",
        "-x~uniform(-50, 50, precision=10)",
        "--test-env",
        "--experiment-id",
        "{exp.id}",
        "--experiment-name",
        "{exp.name}",
        "--experiment-version",
        "{exp.version}",
        "--trial-id",
        "{trial.id}",
        "--working-dir",
        "{trial.working_dir}",
    ]

    orion.core.cli.main(["hunt", "--config", "./orion_config.yaml"] + user_args)

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    exp_id = exp["_id"]
    assert exp["name"] == "voila_voici"
    assert exp["pool_size"] == 1
    assert exp["max_trials"] == 20
    assert exp["max_broken"] == 5
    assert exp["algorithms"] == {
        "gradient_descent": {"learning_rate": 0.1, "dx_tolerance": 1e-5}
    }
    assert "user" in exp["metadata"]
    assert "datetime" in exp["metadata"]
    assert "orion_version" in exp["metadata"]
    assert "user_script" in exp["metadata"]
    assert exp["metadata"]["user_args"] == user_args
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) <= 15
    assert trials[-1].status == "completed"
    trials = list(sorted(trials, key=lambda trial: trial.submit_time))
    for result in trials[-1].results:
        assert result.type != "constraint"
        if result.type == "objective":
            assert abs(result.value - 23.4) < 1e-6
            assert result.name == "example_objective"
        elif result.type == "gradient":
            res = numpy.asarray(result.value)
            assert 0.1 * numpy.sqrt(res.dot(res)) < 1e-4
            assert result.name == "example_gradient"
    params = trials[-1].params
    assert len(params) == 1
    px = params["/x"]
    assert isinstance(px, float)
    assert (px - 34.56789) < 1e-4


def test_demo_with_script_config(storage, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "./black_box_w_config.py",
            "--config",
            "script_config.yaml",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    exp_id = exp["_id"]
    assert exp["name"] == "voila_voici"
    assert exp["pool_size"] == 1
    assert exp["max_trials"] == 20
    assert exp["max_broken"] == 5
    assert exp["algorithms"] == {
        "gradient_descent": {"learning_rate": 0.1, "dx_tolerance": 1e-5}
    }
    assert "user" in exp["metadata"]
    assert "datetime" in exp["metadata"]
    assert "orion_version" in exp["metadata"]
    assert "user_script" in exp["metadata"]
    assert exp["metadata"]["user_args"] == [
        "./black_box_w_config.py",
        "--config",
        "script_config.yaml",
    ]

    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) <= 15
    assert trials[-1].status == "completed"
    trials = list(sorted(trials, key=lambda trial: trial.submit_time))
    for result in trials[-1].results:
        assert result.type != "constraint"
        if result.type == "objective":
            assert abs(result.value - 23.4) < 1e-6
            assert result.name == "example_objective"
        elif result.type == "gradient":
            res = numpy.asarray(result.value)
            assert 0.1 * numpy.sqrt(res.dot(res)) < 1e-4
            assert result.name == "example_gradient"
    params = trials[-1].params
    assert len(params) == 1
    px = params["/x"]
    assert isinstance(px, float)
    assert (px - 34.56789) < 1e-4


def test_demo_with_python_and_script(storage, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "python",
            "black_box_w_config.py",
            "--config",
            "script_config.yaml",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    exp_id = exp["_id"]
    assert exp["name"] == "voila_voici"
    assert exp["pool_size"] == 1
    assert exp["max_trials"] == 20
    assert exp["max_broken"] == 5
    assert exp["algorithms"] == {
        "gradient_descent": {"learning_rate": 0.1, "dx_tolerance": 1e-5}
    }
    assert "user" in exp["metadata"]
    assert "datetime" in exp["metadata"]
    assert "orion_version" in exp["metadata"]
    assert "user_script" in exp["metadata"]
    assert exp["metadata"]["user_args"] == [
        "python",
        "black_box_w_config.py",
        "--config",
        "script_config.yaml",
    ]

    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) <= 15
    assert trials[-1].status == "completed"
    trials = list(sorted(trials, key=lambda trial: trial.submit_time))
    for result in trials[-1].results:
        assert result.type != "constraint"
        if result.type == "objective":
            assert abs(result.value - 23.4) < 1e-6
            assert result.name == "example_objective"
        elif result.type == "gradient":
            res = numpy.asarray(result.value)
            assert 0.1 * numpy.sqrt(res.dot(res)) < 1e-4
            assert result.name == "example_gradient"
    params = trials[-1].params
    assert len(params) == 1
    px = params["/x"]
    assert isinstance(px, float)
    assert (px - 34.56789) < 1e-4


def test_demo_inexecutable_script(storage, monkeypatch, capsys):
    """Test error message when user script is not executable."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    script = tempfile.NamedTemporaryFile()
    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            script.name,
            "--config",
            "script_config.yaml",
        ]
    )

    captured = capsys.readouterr().err
    assert "User script is not executable" in captured


def test_demo_four_workers(storage, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    processes = []
    for _ in range(4):
        process = subprocess.Popen(
            [
                "orion",
                "hunt",
                "-n",
                "four_workers_demo",
                "--config",
                "./orion_config_random.yaml",
                "--max-trials",
                "20",
                "./black_box.py",
                "-x~norm(34, 3)",
            ]
        )
        processes.append(process)

    for process in processes:
        rcode = process.wait()
        assert rcode == 0

    exp = list(storage.fetch_experiments({"name": "four_workers_demo"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    exp_id = exp["_id"]
    assert exp["name"] == "four_workers_demo"
    assert exp["pool_size"] == 2
    assert exp["max_trials"] == 20
    assert exp["max_broken"] == 5
    assert exp["algorithms"] == {"random": {"seed": 2}}
    assert "user" in exp["metadata"]
    assert "datetime" in exp["metadata"]
    assert "orion_version" in exp["metadata"]
    assert "user_script" in exp["metadata"]
    assert exp["metadata"]["user_args"] == ["./black_box.py", "-x~norm(34, 3)"]

    trials = list(storage.fetch_trials(uid=exp_id))
    status = defaultdict(int)
    for trial in trials:
        status[trial.status] += 1
    assert status["completed"] == 20
    assert status["new"] < 5
    params = trials[-1].params
    assert len(params) == 1
    px = params["/x"]
    assert isinstance(px, float)


def test_workon():
    """Test scenario having a configured experiment already setup."""
    name = "voici_voila"
    config = {"name": name}
    config["algorithms"] = {"random": {"seed": 1}}
    config["pool_size"] = 1
    config["max_trials"] = 50
    config["exp_max_broken"] = 5
    config["user_args"] = [
        os.path.abspath(os.path.join(os.path.dirname(__file__), "black_box.py")),
        "-x~uniform(-50, 50, precision=None)",
    ]

    with OrionState():
        experiment = experiment_builder.build_from_args(config)

        workon(
            experiment,
            n_workers=2,
            max_trials=10,
            max_broken=5,
            max_idle_time=20,
            heartbeat=20,
            user_script_config="config",
            interrupt_signal_code=120,
            ignore_code_changes=True,
            executor="joblib",
            executor_configuration={"backend": "threading"},
        )

        storage = get_storage()

        exp = list(storage.fetch_experiments({"name": name}))
        assert len(exp) == 1
        exp = exp[0]
        assert "_id" in exp
        assert exp["name"] == name
        assert exp["pool_size"] == 1
        assert exp["max_trials"] == 50
        assert exp["max_broken"] == 5
        assert exp["algorithms"] == {"random": {"seed": 1}}
        assert "user" in exp["metadata"]
        assert "datetime" in exp["metadata"]
        assert "user_script" in exp["metadata"]
        assert exp["metadata"]["user_args"] == config["user_args"]

        trials = experiment.fetch_trials_by_status("completed")
        assert len(trials) <= 22
        trials = list(sorted(trials, key=lambda trial: trial.submit_time))
        assert trials[-1].status == "completed"
        params = trials[-1].params
        assert len(params) == 1
        px = params["/x"]
        assert isinstance(px, float)
        assert (px - 34.56789) < 5


def test_stress_unique_folder_creation(storage, monkeypatch, tmpdir, capfd):
    """Test integration with a possible framework that needs to create
    unique directories per trial.
    """
    # XXX: return and complete test when there is a way to control random
    # seed of Oríon
    how_many = 2
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(
        [
            "hunt",
            "--max-trials={}".format(how_many),
            "--pool-size=1",
            "--name=lalala",
            "--config",
            "./stress_gradient.yaml",
            "./dir_per_trial.py",
            "--dir={}".format(str(tmpdir)),
            "--other-name",
            "{exp.name}",
            "--name",
            "{trial.hash_name}",
            "-x~gaussian(30, 10)",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "lalala"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    exp_id = exp["_id"]

    # For contingent broken trials, which in this test means that a existing
    # directory was attempted to be created, it means that it's not md5 or
    # bad hash creation to blame, but the finite precision of the floating
    # point representation. Specifically, it seems that gradient descent
    # is able to reach such levels of precision jumping around the minimum
    # (notice that an appropriate learning rate was selected in this stress
    # test to create underdamped behaviour), that it begins to suggest same
    # things from the past. This is intended to be shown with the assertions
    # in the for-loop below.
    trials_c = list(storage.fetch_trials(uid=exp_id, where={"status": "completed"}))
    list_of_cx = [trial.params["/x"] for trial in trials_c]
    trials_b = list(storage.fetch_trials(uid=exp_id, where={"status": "broken"}))
    list_of_bx = [trial.params["/x"] for trial in trials_b]
    for bx in list_of_bx:
        assert bx in list_of_cx

    # ``exp.name`` has been delivered correctly (next 2 assertions)
    assert len(os.listdir(str(tmpdir))) == 1
    # Also, because of the way the demo gradient descent works `how_many` trials
    # can be completed
    assert len(os.listdir(str(tmpdir.join("lalala")))) == how_many
    assert len(trials_c) == how_many
    capfd.readouterr()  # Suppress fd level 1 & 2


def test_working_dir_argument_cmdline(storage, monkeypatch, tmp_path):
    """Check that a permanent directory is used instead of tmpdir"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    path = str(tmp_path) + "/test"
    assert not os.path.exists(path)
    orion.core.cli.main(
        [
            "hunt",
            "-n",
            "allo",
            "--working-dir",
            path,
            "--max-trials",
            "2",
            "--config",
            "./database_config.yaml",
            "./black_box.py",
            "-x~uniform(-50,50)",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "allo"}))[0]
    assert exp["working_dir"] == path
    assert os.path.exists(path)
    assert os.listdir(path)

    shutil.rmtree(path)


def test_tmpdir_is_deleted(storage, monkeypatch, tmp_path):
    """Check that temporary directory is deletid tmpdir"""
    tmp_path = os.path.join(tempfile.gettempdir(), "orion")
    if os.path.exists(tmp_path):
        shutil.rmtree(tmp_path)

    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(
        [
            "hunt",
            "-n",
            "allo",
            "--max-trials",
            "2",
            "--config",
            "./database_config.yaml",
            "./black_box.py",
            "-x~uniform(-50,50)",
        ]
    )

    assert not os.listdir(tmp_path)


def test_working_dir_argument_config(storage, monkeypatch):
    """Check that workning dir argument is handled properly"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    dir_path = os.path.join("orion", "test")
    if os.path.exists(dir_path):
        shutil.rmtree(dir_path)

    orion.core.cli.main(
        [
            "hunt",
            "-n",
            "allo",
            "--max-trials",
            "2",
            "--config",
            "./working_dir_config.yaml",
            "./black_box.py",
            "-x~uniform(-50,50)",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "allo"}))[0]
    assert exp["working_dir"] == dir_path
    assert os.path.exists(dir_path)
    assert os.listdir(dir_path)

    shutil.rmtree(dir_path)


def test_run_with_name_only(storage, monkeypatch):
    """Test hunt can be executed with experiment name only"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "--config",
            "./orion_config_random.yaml",
            "./black_box.py",
            "-x~uniform(-50, 50)",
        ]
    )

    orion.core.cli.main(
        ["hunt", "--max-trials", "20", "--config", "./orion_config_random.yaml"]
    )

    exp = list(storage.fetch_experiments({"name": "demo_random_search"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) == 20


def test_run_with_name_only_with_trailing_whitespace(storage, monkeypatch):
    """Test hunt can be executed with experiment name and trailing whitespace"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(
        [
            "hunt",
            "--init-only",
            "--config",
            "./orion_config_random.yaml",
            "./black_box.py",
            "-x~uniform(-50, 50)",
        ]
    )

    orion.core.cli.main(
        ["hunt", "--max-trials", "20", "--config", "./orion_config_random.yaml", ""]
    )

    exp = list(storage.fetch_experiments({"name": "demo_random_search"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) == 20


@pytest.mark.parametrize("strategy", ["MaxParallelStrategy", "MeanParallelStrategy"])
def test_run_with_parallel_strategy(storage, monkeypatch, strategy):
    """Test hunt can be executed with max parallel strategies"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    with open("strategy_config.yaml") as f:
        config = yaml.safe_load(f.read())

    config_file = "{}_strategy_config.yaml".format(strategy)

    with open(config_file, "w") as f:
        config["producer"]["strategy"] = strategy
        f.write(yaml.dump(config))

    orion.core.cli.main(
        [
            "hunt",
            "--max-trials",
            "20",
            "--pool-size",
            "1",
            "--config",
            config_file,
            "./black_box.py",
            "-x~uniform(-50, 50)",
        ]
    )

    os.remove(config_file)

    exp = list(storage.fetch_experiments({"name": "strategy_demo"}))
    assert len(exp) == 1
    exp = exp[0]
    assert exp["producer"]["strategy"] == {strategy: {"default_result": float("inf")}}
    assert "_id" in exp
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) == 20


def test_worker_trials(storage, monkeypatch):
    """Test number of trials executed is limited based on worker-trials"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    assert len(list(storage.fetch_experiments({"name": "demo_random_search"}))) == 0

    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config_random.yaml",
            "--pool-size",
            "1",
            "--worker-trials",
            "0",
            "./black_box.py",
            "-x~uniform(-50, 50)",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "demo_random_search"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    exp_id = exp["_id"]

    def n_completed():
        return len(
            list(storage._fetch_trials({"experiment": exp_id, "status": "completed"}))
        )

    assert n_completed() == 0

    # Test only executes 2 trials
    orion.core.cli.main(
        ["hunt", "--name", "demo_random_search", "--worker-trials", "2"]
    )

    assert n_completed() == 2

    # Test only executes 3 more trials
    orion.core.cli.main(
        ["hunt", "--name", "demo_random_search", "--worker-trials", "3"]
    )

    assert n_completed() == 5

    # Test that max-trials has precedence over worker-trials
    orion.core.cli.main(
        [
            "hunt",
            "--name",
            "demo_random_search",
            "--worker-trials",
            "5",
            "--max-trials",
            "6",
        ]
    )

    assert n_completed() == 6


@pytest.mark.usefixtures("storage")
def test_resilience(monkeypatch):
    """Test if Oríon stops after enough broken trials."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    MAX_BROKEN = 3
    orion.core.config.worker.max_broken = 3

    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config_random.yaml",
            "./broken_box.py",
            "-x~uniform(-50, 50)",
        ]
    )

    exp = experiment_builder.build(name="demo_random_search")
    assert len(exp.fetch_trials_by_status("broken")) == MAX_BROKEN


@pytest.mark.usefixtures("storage")
def test_demo_with_shutdown_quickly(monkeypatch):
    """Check simple pipeline with random search is reasonably fast."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    monkeypatch.setattr(orion.core.config.worker, "heartbeat", 120)

    process = subprocess.Popen(
        [
            "orion",
            "hunt",
            "--config",
            "./orion_config_random.yaml",
            "--max-trials",
            "10",
            "./black_box.py",
            "-x~uniform(-50, 50)",
        ]
    )

    assert process.wait(timeout=40) == 0


def test_demo_with_nondefault_config_keyword(storage, monkeypatch):
    """Check that the user script configuration file is correctly used with a new keyword."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.config.worker.user_script_config = "configuration"
    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config_other.yaml",
            "./black_box_w_config_other.py",
            "--configuration",
            "script_config.yaml",
        ]
    )

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    assert len(exp) == 1
    exp = exp[0]
    assert "_id" in exp
    exp_id = exp["_id"]
    assert exp["name"] == "voila_voici"
    assert exp["pool_size"] == 1
    assert exp["max_trials"] == 20
    assert exp["algorithms"] == {
        "gradient_descent": {"learning_rate": 0.1, "dx_tolerance": 1e-5}
    }
    assert "user" in exp["metadata"]
    assert "datetime" in exp["metadata"]
    assert "orion_version" in exp["metadata"]
    assert "user_script" in exp["metadata"]
    assert exp["metadata"]["user_args"] == [
        "./black_box_w_config_other.py",
        "--configuration",
        "script_config.yaml",
    ]

    trials = list(storage.fetch_trials(uid=exp_id))
    assert len(trials) <= 15
    assert trials[-1].status == "completed"
    trials = list(sorted(trials, key=lambda trial: trial.submit_time))
    for result in trials[-1].results:
        assert result.type != "constraint"
        if result.type == "objective":
            assert abs(result.value - 23.4) < 1e-6
            assert result.name == "example_objective"
        elif result.type == "gradient":
            res = numpy.asarray(result.value)
            assert 0.1 * numpy.sqrt(res.dot(res)) < 1e-4
            assert result.name == "example_gradient"
    params = trials[-1].params
    assert len(params) == 1
    px = params["/x"]
    assert isinstance(px, float)
    assert (px - 34.56789) < 1e-4

    orion.core.config.worker.user_script_config = "config"


def test_demo_precision(storage, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main(
        [
            "hunt",
            "--config",
            "./orion_config.yaml",
            "--max-trials",
            "2",
            "./black_box.py",
        ]
        + user_args
    )

    exp = list(storage.fetch_experiments({"name": "voila_voici"}))
    exp = exp[0]
    exp_id = exp["_id"]
    trials = list(storage.fetch_trials(uid=exp_id))
    trials = list(sorted(trials, key=lambda trial: trial.submit_time))
    params = trials[-1].params
    value = params["/x"]

    assert value == float(numpy.format_float_scientific(value, precision=4))


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_debug_mode(monkeypatch):
    """Test debug mode."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = ["-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main(
        [
            "--debug",
            "hunt",
            "--config",
            "./orion_config.yaml",
            "--max-trials",
            "2",
            "./black_box.py",
        ]
        + user_args
    )

    storage = get_storage()

    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, EphemeralDB)


def test_no_args(capsys):
    """Test that help is printed when no args are given."""
    with pytest.raises(SystemExit):
        orion.core.cli.main([])

    captured = capsys.readouterr().out

    assert "usage:" in captured
    assert "Traceback" not in captured



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the db commands."""
import pytest

import orion.core.cli


def test_no_args(capsys):
    """Test that help is printed when no args are given."""
    with pytest.raises(SystemExit):
        orion.core.cli.main(["db"])

    captured = capsys.readouterr().out

    assert "usage:" in captured
    assert "Traceback" not in captured



# -*- coding: utf-8 -*-
"""Perform functional tests for the REST endpoint `/`"""


def test_runtime_summary(client):
    """Tests if the instance's meta-summary is present"""
    response = client.simulate_get("/")

    assert response.json["orion"]
    assert response.json["server"] == "gunicorn"
    assert response.json["database"] == "EphemeralDB"
    assert response.status == "200 OK"



# -*- coding: utf-8 -*-
"""
Base Search Algorithm
=====================

Formulation of a general search algorithm with respect to some objective.
Algorithm implementations must inherit from `orion.algo.base.OptimizationAlgorithm`.

"""
import copy
import hashlib
import logging
from abc import ABCMeta, abstractmethod

from orion.algo.space import Fidelity
from orion.core.utils import Factory

log = logging.getLogger(__name__)


def infer_trial_id(point):
    """Compute a hashing of a point"""
    return hashlib.md5(str(list(point)).encode("utf-8")).hexdigest()


# pylint: disable=too-many-public-methods
class BaseAlgorithm(object, metaclass=ABCMeta):
    """Base class describing what an algorithm can do.

    Parameters
    ----------
    space : `orion.algo.space.Space`
       Definition of a problem's parameter space.
    kwargs : dict
       Tunable elements of a particular algorithm, a dictionary from
       hyperparameter names to values.

    Notes
    -----
    We are using the No Free Lunch theorem's [1]_[3]_ formulation of an
    `BaseAlgorithm`.
    We treat it as a part of a procedure which in each iteration suggests a
    sample of the parameter space of the problem as a candidate solution and observes
    the results of its evaluation.

    **Developer Note**: Each algorithm's complete specification, i.e.  implementation of its methods
    and parameters of its own, lies in a separate concrete algorithm class, which must be an
    **immediate** subclass of `BaseAlgorithm`. [The reason for this is current implementation of
    `orion.core.utils.Factory` metaclass which uses `BaseAlgorithm.__subclasses__()`.] Second, one
    must declare an algorithm's own parameters (tunable elements which could be set by
    configuration). This is done by passing them to `BaseAlgorithm.__init__()` by calling Python's
    super with a `Space` object as a positional argument plus algorithm's own parameters as keyword
    arguments. The keys of the keyword arguments passed to `BaseAlgorithm.__init__()` are interpreted
    as the algorithm's parameter names. So for example, a subclass could be as simple as this
    (regarding the logistics, not an actual algorithm's implementation):

    Examples
    --------
    .. code-block:: python
       :linenos:
       :emphasize-lines: 7

       from orion.algo.base import BaseAlgorithm
       from orion.algo.space import (Integer, Space)

       class MySimpleAlgo(BaseAlgorithm):

           def __init__(self, space, multiplier=1, another_param="a string param"):
               super().__init__(space, multiplier=multiplier, another_param=another_param)

           def suggest(self, num=1):
               print(self.another_param)
               return list(map(lambda x: tuple(map(lambda y: self.multiplier * y, x)),
                               self.space.sample(num)))

           def observe(self, points, results):
               pass

       dim = Integer('named_param', 'norm', 3, 2, shape=(2, 3))
       s = Space()
       s.register(dim)

       algo = MySimpleAlgo(s, 2, "I am just sampling!")
       algo.suggest()

    References
    ----------
    .. [1] D. H. Wolpert and W. G. Macready, “No Free Lunch Theorems for Optimization,”
       IEEE Transactions on Evolutionary Computation, vol. 1, no. 1, pp. 67–82, Apr. 1997.
    .. [2] W. G. Macready and D. H. Wolpert, “What Makes An Optimization Problem Hard?,”
       Complexity, vol. 1, no. 5, pp. 40–46, 1996.
    .. [3] D. H. Wolpert and W. G. Macready, “No Free Lunch Theorems for Search,”
       Technical Report SFI-TR-95-02-010, Santa Fe Institute, 1995.

    """

    requires_type = None
    requires_shape = None
    requires_dist = None

    def __init__(self, space, **kwargs):
        log.debug(
            "Creating Algorithm object of %s type with parameters:\n%s",
            type(self).__name__,
            kwargs,
        )
        self._trials_info = {}  # Stores Unique Trial -> Result
        self._space = space
        self._param_names = list(kwargs.keys())
        # Instantiate tunable parameters of an algorithm
        for varname, param in kwargs.items():
            # Check if tunable element is another algorithm
            if isinstance(param, dict) and len(param) == 1:
                subalgo_type = list(param)[0]
                subalgo_kwargs = param[subalgo_type]
                if isinstance(subalgo_kwargs, dict):
                    param = OptimizationAlgorithm(subalgo_type, space, **subalgo_kwargs)
            elif (
                isinstance(param, str) and param.lower() in OptimizationAlgorithm.types
            ):
                # pylint: disable=too-many-function-args
                param = OptimizationAlgorithm(param, space)
            elif varname == "seed":
                self.seed_rng(param)

            setattr(self, varname, param)

    def seed_rng(self, seed):
        """Seed the state of the random number generator.

        :param seed: Integer seed for the random number generator.

        .. note:: This methods does nothing if the algorithm is deterministic.
        """
        pass

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        return {"_trials_info": copy.deepcopy(self._trials_info)}

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict

        :param state_dict: Dictionary representing state of an algorithm
        """
        self._trials_info = state_dict.get("_trials_info")

    def format_point(self, point):
        """Format point based on space transformations

        This will apply the reverse transformation on the point and then
        transform it again.

        Some transformations are lossy and thus the points suggested by the algorithm could
        be different when returned to `observe`. Using `format_point` makes it possible
        for the algorithm to see the final version of the point after back and forth
        transformations. This way it can recognise the point in `observe` and also
        avoid duplicates that would have gone unnoticed during suggestion.

        Parameters
        ----------
        point : tuples of array-likes
            Points from a `orion.algo.space.Space`.
        """

        point = tuple(point)
        if hasattr(self.space, "transform"):
            point = self.space.transform(self.space.reverse(point))

        return point

    def get_id(self, point, ignore_fidelity=False):
        """Compute a unique hash for a point based on params

        Parameters
        ----------
        point : tuples of array-likes
            Points from a `orion.algo.space.Space`.
        ignore_fidelity: bool, optional
            If True, the fidelity dimension is ignored when computing a unique hash for
            the trial. Defaults to False.
        """
        # Apply transforms and reverse to see data as it would come from DB
        # (Some transformations looses some info. ex: Precision transformation)
        point = list(self.format_point(point))

        if ignore_fidelity:
            non_fidelity_dims = point[0 : self.fidelity_index]
            non_fidelity_dims.extend(point[self.fidelity_index + 1 :])
            point = non_fidelity_dims

        return hashlib.md5(str(point).encode("utf-8")).hexdigest()

    @property
    def fidelity_index(self):
        """Compute the index of the point where fidelity is.

        Returns None if there is no fidelity dimension.
        """

        def _is_fidelity(dim):
            return dim.type == "fidelity"

        fidelity_index = [
            i for i, dim in enumerate(self.space.values()) if _is_fidelity(dim)
        ]
        if fidelity_index:
            return fidelity_index[0]

        return None

    @abstractmethod
    def suggest(self, num):
        """Suggest a `num` of new sets of parameters.

        Parameters
        ----------
        num: int
            Number of points to suggest. The algorithm may return less than the number of points
            requested.

        Returns
        -------
        list of points or None
            A list of lists representing points suggested by the algorithm. The algorithm may opt
            out if it cannot make a good suggestion at the moment (it may be waiting for other
            trials to complete), in which case it will return None.

        Notes
        -----
        New parameters must be compliant with the problem's domain `orion.algo.space.Space`.

        """
        pass

    def observe(self, points, results):
        """Observe the `results` of the evaluation of the `points` in the
        process defined in user's script.

        Parameters
        ----------
        points : list of tuples of array-likes
           Points from a `orion.algo.space.Space`.
        results : list of dicts
           Contains the result of an evaluation; partial information about the
           black-box function at each point in `params`.

        Result
        ------
        objective : numeric
           Evaluation of this problem's objective function.
        gradient : 1D array-like, optional
           Contains values of the derivatives of the `objective` function
           with respect to `params`.
        constraint : list of numeric, optional
           List of constraints expression evaluation which must be greater
           or equal to zero by the problem's definition.

        """
        for point, result in zip(points, results):
            if not self.has_observed(point):
                self.register(point, result)

    def register(self, point, result=None):
        """Save the point as one suggested or observed by the algorithm

        Parameters
        ----------
        point : array-likes
           Point from a `orion.algo.space.Space`.
        result : dict or None, optional
           The result of an evaluation; partial information about the
           black-box function at each point in `params`.
           None is suggested and not yet completed.

        """
        self._trials_info[self.get_id(point)] = (point, result)

    @property
    def n_suggested(self):
        """Number of trials suggested by the algorithm"""
        return len(self._trials_info)

    @property
    def n_observed(self):
        """Number of completed trials observed by the algorithm"""
        return sum(bool(point[1] is not None) for point in self._trials_info.values())

    def has_suggested(self, point):
        """Whether the algorithm has suggested a given point.

        Parameters
        ----------
        point : tuples of array-likes
           Points from a `orion.algo.space.Space`.

        Returns
        -------
        bool
            True if the point was suggested by the algo, False otherwise.

        """
        return self.get_id(point) in self._trials_info

    def has_observed(self, point):
        """Whether the algorithm has observed a given point objective.

        This only counts observed completed trials.

        Parameters
        ----------
        point : tuples of array-likes
            Points from a `orion.algo.space.Space`.

        Returns
        -------
        bool
            True if the point's objective was observed by the algo, False otherwise.

        """

        trial_id = self.get_id(point)
        return (
            trial_id in self._trials_info and self._trials_info[trial_id][1] is not None
        )

    @property
    def is_done(self):
        """Whether the algorithm is done and will not make further suggestions.

        Return True, if an algorithm holds that there can be no further improvement.
        By default, the cardinality of the specified search space will be used to check
        if all possible sets of parameters has been tried.
        """
        if self.n_suggested >= self.space.cardinality:
            return True

        if self.n_observed >= getattr(self, "max_trials", float("inf")):
            return True

        return False

    def score(self, point):  # pylint:disable=no-self-use,unused-argument
        """Allow algorithm to evaluate `point` based on a prediction about
        this parameter set's performance.

        By default, return the same score any parameter (no preference).

        :returns: A subjective measure of expected perfomance.
        :rtype: float

        """
        return 0

    def judge(self, point, measurements):  # pylint:disable=no-self-use,unused-argument
        """Inform an algorithm about online `measurements` of a running trial.

        :param point: A tuple which specifies the values of the (hyper)parameters
           used to execute user's script with.

        This method is to be used as a callback in a client-server communication
        between user's script and a orion's worker using a `BaseAlgorithm`.
        Data returned from this method must be serializable and will be used as
        a response to the running environment. Default response is None.

        .. note:: Calling algorithm to `judge` a `point` based on its online
           `measurements` will effectively change a state in the algorithm (like
           a reinforcement learning agent's hidden state or an automatic early
           stopping mechanism's regression), which it may change the value of
           the property `should_suspend`.

        :returns: None or a serializable dictionary containing named data

        """
        return None

    @property
    def should_suspend(self):
        """Allow algorithm to decide whether a particular running trial is still
        worth to complete its evaluation, based on information provided by the
        `judge` method.

        """
        return False

    @property
    def configuration(self):
        """Return tunable elements of this algorithm in a dictionary form
        appropriate for saving.

        """
        dict_form = dict()
        for attrname in self._param_names:
            if attrname.startswith("_"):  # Do not log _space or others in conf
                continue
            attr = getattr(self, attrname)
            if isinstance(attr, BaseAlgorithm):
                attr = attr.configuration
            dict_form[attrname] = attr

        return {self.__class__.__name__.lower(): dict_form}

    @property
    def space(self):
        """Domain of problem associated with this algorithm's instance."""
        return self._space

    @space.setter
    def space(self, space_):
        """Propagate changes in defined space to possibly nested algorithms."""
        self._space = space_
        for attr in self.__dict__.values():
            if isinstance(attr, BaseAlgorithm):
                attr.space = space_


# pylint: disable=too-few-public-methods,abstract-method
class OptimizationAlgorithm(BaseAlgorithm, metaclass=Factory):
    """Class used to inject dependency on an algorithm implementation."""

    pass



# -*- coding: utf-8 -*-
"""Generic tests for Space"""
from collections import OrderedDict

from orion.algo.space import Categorical, Dimension, Fidelity, Integer, Real, Space


def dim1():
    """Create an example of `orion.algo.space.Dimension`."""
    dim = Real("yolo0", "norm", 0.9, shape=(3, 2))
    return dim


def dim2():
    """Create a second example of `orion.algo.space.Dimension`."""
    probs = (0.1, 0.2, 0.3, 0.4)
    categories = ("asdfa", "2", "3", "4")
    categories = OrderedDict(zip(categories, probs))
    dim2 = Categorical("yolo2", categories)
    return dim2


def dim3():
    """Create an example of integer `orion.algo.space.Dimension`."""
    return Integer("yolo3", "uniform", 3, 7)


def dim4():
    """Create an example of Fidelity `orion.algo.space.Dimension`."""
    return Fidelity("yolo4", 1, 10)


def build_space(dims=None):
    """Create an example `orion.algo.space.Space`."""
    if dims is None:
        dims = [dim1(), dim2(), dim3(), dim4()]
    space = Space()
    for dim in dims:
        space.register(dim)
    return space



"""Example usage and tests for :mod:`orion.core.utils.flatten`."""
from orion.core.utils.flatten import flatten, unflatten


def test_basic():
    """Test basic functionality of flatten"""
    d = {"a": {"b": 2, "c": 3}, "c": {"d": 3, "e": 4}}
    assert flatten(d) == {"a.b": 2, "a.c": 3, "c.d": 3, "c.e": 4}


def test_handle_double_ref():
    """Test proper handling of double references in dicts"""
    a = {"b": 2, "c": 3}
    d = {"a": a, "d": {"e": a}}
    assert flatten(d) == {"a.b": 2, "a.c": 3, "d.e.b": 2, "d.e.c": 3}


def test_unflatten():
    """Test than unflatten(flatten(x)) is idempotent"""
    a = {"b": 2, "c": 3}
    d = {"a": a, "d": {"e": a}}
    assert unflatten(flatten(d)) == d



#!/usr/bin/env python
"""Common fixtures and utils for tests."""

import copy
import datetime
import getpass
import os

import pytest

import orion.core.io.experiment_builder as experiment_builder
import orion.core.utils.backward as backward
from orion.algo.space import Categorical, Integer, Real, Space
from orion.core.evc import conflicts
from orion.core.io.convert import JSONConverter, YAMLConverter
from orion.core.io.space_builder import DimensionBuilder
from orion.core.worker.trial import Trial
from orion.testing import MockDatetime, default_datetime

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
YAML_SAMPLE = os.path.join(TEST_DIR, "sample_config.yml")
YAML_DIFF_SAMPLE = os.path.join(TEST_DIR, "sample_config_diff.yml")
JSON_SAMPLE = os.path.join(TEST_DIR, "sample_config.json")
UNKNOWN_SAMPLE = os.path.join(TEST_DIR, "sample_config.txt")
UNKNOWN_TEMPLATE = os.path.join(TEST_DIR, "sample_config_template.txt")


@pytest.fixture(scope="session")
def yaml_sample_path():
    """Return path with a yaml sample file."""
    return os.path.abspath(YAML_SAMPLE)


@pytest.fixture(scope="session")
def yaml_diff_sample_path():
    """Return path with a different yaml sample file."""
    return os.path.abspath(YAML_DIFF_SAMPLE)


@pytest.fixture
def yaml_config(yaml_sample_path):
    """Return a list containing the key and the sample path for a yaml config."""
    return ["--config", yaml_sample_path]


@pytest.fixture
def yaml_diff_config(yaml_diff_sample_path):
    """Return a list containing the key and the sample path for a different yaml config."""
    return ["--config", yaml_diff_sample_path]


@pytest.fixture(scope="session")
def json_sample_path():
    """Return path with a json sample file."""
    return JSON_SAMPLE


@pytest.fixture
def json_config(json_sample_path):
    """Return a list containing the key and the sample path for a json config."""
    return ["--config", json_sample_path]


@pytest.fixture(scope="session")
def unknown_type_sample_path():
    """Return path with a sample file of unknown configuration filetype."""
    return UNKNOWN_SAMPLE


@pytest.fixture(scope="session")
def unknown_type_template_path():
    """Return path with a template file of unknown configuration filetype."""
    return UNKNOWN_TEMPLATE


@pytest.fixture(scope="session")
def some_sample_path():
    """Return path with a sample file of unknown configuration filetype."""
    return os.path.join(TEST_DIR, "some_sample_config.txt")


@pytest.fixture
def some_sample_config(some_sample_path):
    """Return a list containing the key and the sample path for some config."""
    return ["--config", some_sample_path]


@pytest.fixture(scope="session")
def yaml_converter():
    """Return a yaml converter."""
    return YAMLConverter()


@pytest.fixture(scope="session")
def json_converter():
    """Return a json converter."""
    return JSONConverter()


@pytest.fixture(scope="module")
def space():
    """Construct a simple space with every possible kind of Dimension."""
    space = Space()
    categories = {"asdfa": 0.1, 2: 0.2, 3: 0.3, 4: 0.4}
    dim = Categorical("yolo", categories, shape=2)
    space.register(dim)
    dim = Integer("yolo2", "uniform", -3, 6)
    space.register(dim)
    dim = Real("yolo3", "alpha", 0.9)
    space.register(dim)
    return space


@pytest.fixture(scope="module")
def hierarchical_space():
    """Construct a space with hierarchical Dimensions."""
    space = Space()
    categories = {"asdfa": 0.1, 2: 0.2, 3: 0.3, 4: 0.4}
    dim = Categorical("yolo.first", categories, shape=2)
    space.register(dim)
    dim = Integer("yolo.second", "uniform", -3, 6)
    space.register(dim)
    dim = Real("yoloflat", "alpha", 0.9)
    space.register(dim)
    return space


@pytest.fixture(scope="module")
def fixed_suggestion():
    """Return the same tuple/sample from a possible space."""
    return (("asdfa", 2), 0, 3.5)


@pytest.fixture()
def with_user_tsirif(monkeypatch):
    """Make ``getpass.getuser()`` return ``'tsirif'``."""
    monkeypatch.setattr(getpass, "getuser", lambda: "tsirif")


@pytest.fixture()
def with_user_bouthilx(monkeypatch):
    """Make ``getpass.getuser()`` return ``'bouthilx'``."""
    monkeypatch.setattr(getpass, "getuser", lambda: "bouthilx")


@pytest.fixture()
def with_user_dendi(monkeypatch):
    """Make ``getpass.getuser()`` return ``'dendi'``."""
    monkeypatch.setattr(getpass, "getuser", lambda: "dendi")


@pytest.fixture()
def random_dt(monkeypatch):
    """Make ``datetime.datetime.utcnow()`` return an arbitrary date."""
    monkeypatch.setattr(datetime, "datetime", MockDatetime)
    return default_datetime()


dendi_exp_config = dict(
    name="supernaedo2-dendi",
    space={
        "/decoding_layer": f"choices(['rnn', 'lstm_with_attention', 'gru'])",
        "/encoding_layer": f"choices(['rnn', 'lstm', 'gru'])",
    },
    metadata={
        "user": "dendi",
        "orion_version": "XYZ",
        "VCS": {
            "type": "git",
            "is_dirty": False,
            "HEAD_sha": "test",
            "active_branch": None,
            "diff_sha": "diff",
        },
    },
    version=1,
    pool_size=1,
    max_trials=1000,
    working_dir="",
    algorithms={"dumbalgo": {}},
    producer={"strategy": "NoParallelStrategy"},
)


dendi_base_trials = [
    {
        "status": "completed",
        "worker": 12512301,
        "submit_time": MockDatetime(2017, 11, 22, 23),
        "start_time": None,
        "end_time": MockDatetime(2017, 11, 22, 23),
        "results": [{"name": None, "type": "objective", "value": 3}],
        "params": [
            {"name": "/decoding_layer", "type": "categorical", "value": "rnn"},
            {"name": "/encoding_layer", "type": "categorical", "value": "lstm"},
        ],
        "parents": [],
    },
    {
        "status": "completed",
        "worker": 23415151,
        "submit_time": MockDatetime(2017, 11, 23, 0),
        "start_time": None,
        "end_time": MockDatetime(2017, 11, 23, 0),
        "results": [
            {"name": "yolo", "type": "objective", "value": 10},
            {"name": "contra", "type": "constraint", "value": 1.2},
            {"name": "naedw_grad", "type": "gradient", "value": [5, 3]},
        ],
        "params": [
            {
                "name": "/decoding_layer",
                "type": "categorical",
                "value": "lstm_with_attention",
            },
            {"name": "/encoding_layer", "type": "categorical", "value": "gru"},
        ],
        "parents": [],
    },
    {
        "status": "completed",
        "worker": 1251231,
        "submit_time": MockDatetime(2017, 11, 22, 23),
        "start_time": None,
        "end_time": MockDatetime(2017, 11, 22, 22),
        "results": [
            {"name": None, "type": "objective", "value": 2},
            {"name": "naedw_grad", "type": "gradient", "value": [-0.1, 2]},
        ],
        "params": [
            {"name": "/decoding_layer", "type": "categorical", "value": "rnn"},
            {"name": "/encoding_layer", "type": "categorical", "value": "rnn"},
        ],
        "parents": [],
    },
    {
        "status": "new",
        "worker": None,
        "submit_time": MockDatetime(2017, 11, 23, 1),
        "start_time": None,
        "end_time": None,
        "results": [{"name": None, "type": "objective", "value": None}],
        "params": [
            {"name": "/decoding_layer", "type": "categorical", "value": "rnn"},
            {"name": "/encoding_layer", "type": "categorical", "value": "gru"},
        ],
        "parents": [],
    },
    {
        "status": "new",
        "worker": None,
        "submit_time": MockDatetime(2017, 11, 23, 2),
        "start_time": None,
        "end_time": None,
        "results": [{"name": None, "type": "objective", "value": None}],
        "params": [
            {
                "name": "/decoding_layer",
                "type": "categorical",
                "value": "lstm_with_attention",
            },
            {"name": "/encoding_layer", "type": "categorical", "value": "rnn"},
        ],
        "parents": [],
    },
    {
        "status": "interrupted",
        "worker": None,
        "submit_time": MockDatetime(2017, 11, 23, 3),
        "start_time": MockDatetime(2017, 11, 23, 3),
        "end_time": None,
        "results": [{"name": None, "type": "objective", "value": None}],
        "params": [
            {
                "name": "/decoding_layer",
                "type": "categorical",
                "value": "lstm_with_attention",
            },
            {"name": "/encoding_layer", "type": "categorical", "value": "lstm"},
        ],
        "parents": [],
    },
    {
        "status": "suspended",
        "worker": None,
        "submit_time": MockDatetime(2017, 11, 23, 4),
        "start_time": MockDatetime(2017, 11, 23, 4),
        "end_time": None,
        "results": [{"name": None, "type": "objective", "value": None}],
        "params": [
            {"name": "/decoding_layer", "type": "categorical", "value": "gru"},
            {"name": "/encoding_layer", "type": "categorical", "value": "lstm"},
        ],
        "parents": [],
    },
]


@pytest.fixture()
def hacked_exp(with_user_dendi, random_dt, storage):
    """Return an `Experiment` instance to find trials in fake database."""
    storage.create_experiment(dendi_exp_config)
    exp = experiment_builder.build(name=dendi_exp_config["name"])
    storage._db.write(
        "trials",
        [Trial(experiment=exp.id, **trial).to_dict() for trial in dendi_base_trials],
    )
    return exp


###
# Fixtures for EVC tests using conflicts, present in both ./evc and ./io.
# Note: Refactoring the EVC out of orion's core should take care of getting those
#       fixtures out of general conftest.py
###


@pytest.fixture
def new_config():
    """Generate a new experiment configuration"""
    user_script = "tests/functional/demo/black_box.py"
    config = dict(
        name="test",
        algorithms="fancy",
        version=1,
        metadata={
            "VCS": "to be changed",
            "user_script": user_script,
            "user_args": [user_script, "--new~normal(0,2)", "--changed~normal(0,2)"],
            "user": "some_user_name",
            "orion_version": "UVW",
        },
    )

    backward.populate_space(config)

    return config


@pytest.fixture
def old_config(storage):
    """Generate an old experiment configuration"""
    user_script = "tests/functional/demo/black_box.py"
    config = dict(
        name="test",
        algorithms="random",
        version=1,
        metadata={
            "VCS": {
                "type": "git",
                "is_dirty": False,
                "HEAD_sha": "test",
                "active_branch": None,
                "diff_sha": "diff",
            },
            "user_script": user_script,
            "user_args": [
                user_script,
                "--missing~uniform(-10,10)",
                "--changed~uniform(-10,10)",
            ],
            "user": "some_user_name",
            "orion_version": "XYZ",
        },
    )

    backward.populate_space(config)

    storage.create_experiment(config)
    return config


@pytest.fixture
def new_dimension_conflict(old_config, new_config):
    """Generate a new dimension conflict for new experiment configuration"""
    name = "new"
    prior = "normal(0, 2)"
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.NewDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def new_dimension_with_default_conflict(old_config, new_config):
    """Generate a new dimension conflict with default value for new experiment configuration"""
    name = "new"
    prior = "normal(0, 2, default_value=0.001)"
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.NewDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def new_dimension_same_prior_conflict(old_config, new_config):
    """Generate a new dimension conflict with different prior for renaming tests"""
    name = "new"
    prior = "uniform(-10, 10)"
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.NewDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def changed_dimension_conflict(old_config, new_config):
    """Generate a changed dimension conflict"""
    name = "changed"
    old_prior = "uniform(-10, 10)"
    new_prior = "normal(0, 2)"
    dimension = DimensionBuilder().build(name, old_prior)
    return conflicts.ChangedDimensionConflict(
        old_config, new_config, dimension, old_prior, new_prior
    )


@pytest.fixture
def missing_dimension_conflict(old_config, new_config):
    """Generate a missing dimension conflict"""
    name = "missing"
    prior = "uniform(-10, 10)"
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.MissingDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def missing_dimension_with_default_conflict(old_config, new_config):
    """Generate a missing dimension conflict with a default value"""
    name = "missing"
    prior = "uniform(-10, 10, default_value=0.0)"
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.MissingDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def algorithm_conflict(old_config, new_config):
    """Generate an algorithm configuration conflict"""
    return conflicts.AlgorithmConflict(old_config, new_config)


@pytest.fixture
def orion_version_conflict(old_config, new_config):
    """Generate an orion version conflict"""
    return conflicts.OrionVersionConflict(old_config, new_config)


@pytest.fixture
def code_conflict(old_config, new_config):
    """Generate a code conflict"""
    return conflicts.CodeConflict(old_config, new_config)


@pytest.fixture
def cli_conflict(old_config, new_config):
    """Generate a commandline conflict"""
    new_config = copy.deepcopy(new_config)
    new_config["metadata"]["user_args"].append("--some-new=args")
    new_config["metadata"]["user_args"].append("--bool-arg")
    backward.populate_space(new_config)
    return conflicts.CommandLineConflict(old_config, new_config)


@pytest.fixture
def config_conflict(old_config, new_config):
    """Generate a script config conflict"""
    return conflicts.ScriptConfigConflict(old_config, new_config)


@pytest.fixture
def experiment_name_conflict(old_config, new_config):
    """Generate an experiment name conflict"""
    return conflicts.ExperimentNameConflict(old_config, new_config)


@pytest.fixture
def bad_exp_parent_config():
    """Generate a new experiment configuration"""
    config = dict(
        _id="test",
        name="test",
        metadata={
            "user": "corneauf",
            "user_args": ["--x~normal(0,1)"],
            "user_script": "tests/functional/demo/black_box.py",
            "orion_version": "XYZ",
        },
        version=1,
        algorithms="random",
    )

    backward.populate_space(config)

    return config


@pytest.fixture
def bad_exp_child_config(bad_exp_parent_config):
    """Generate a new experiment configuration"""
    config = copy.deepcopy(bad_exp_parent_config)
    config["_id"] = "test2"
    config["refers"] = {"parent_id": "test"}
    config["version"] = 2

    return config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Benchmark definition
======================
"""
import copy
import itertools

from tabulate import tabulate

from orion.client import create_experiment


class Benchmark:
    """
    Benchmark definition

    Parameters
    ----------
    name: str
        Name of the benchmark
    algorithms: list, optional
        Algorithms used for benchmark, and for each algorithm, it can be formats as below:

        - A `str` of the algorithm name
        - A `dict`, with only one key and one value, where key is the algorithm name and value is a dict for the algorithm config.
        - A `dict`, with two keys.

            algorithm: str or dict
                Algorithm name in string or a dict with algorithm configure.
            deterministic: bool, optional
                True if it is a deterministic algorithm, then for each assessment, only one experiment
                will be run for this algorithm.

        Examples:

        >>> ["random", "tpe"]
        >>> ["random", {"tpe": {"seed": 1}}]
        >>> [{"algorithm": "random"}, {"algorithm": {"gridsearch": {"n_values": 50}}, "deterministic": True}]

    targets: list, optional
        Targets for the benchmark, each target will be a dict with two keys.

        assess: list
            Assessment objects
        task: list
            Task objects

    storage: dict, optional
        Configuration of the storage backend.
    """

    def __init__(self, name, algorithms, targets, storage=None):
        self._id = None
        self.name = name
        self.algorithms = algorithms
        self.targets = targets
        self.metadata = {}
        self.storage_config = storage

        self.studies = []

    def setup_studies(self):
        """Setup studies to run for the benchmark.
        Benchmark `algorithms`, together with each `task` and `assessment` combination
        define a study.
        """
        for target in self.targets:
            assessments = target["assess"]
            tasks = target["task"]

            for assess, task in itertools.product(*[assessments, tasks]):
                study = Study(self, self.algorithms, assess, task)
                study.setup_experiments()
                self.studies.append(study)

    def process(self, n_workers=1):
        """Run studies experiment"""
        for study in self.studies:
            study.execute(n_workers)

    def status(self, silent=True):
        """Display benchmark status"""
        total_exp_num = 0
        complete_exp_num = 0
        total_trials = 0
        benchmark_status = []
        for study in self.studies:
            for status in study.status():
                column = dict()
                column["Algorithms"] = status["algorithm"]
                column["Assessments"] = status["assessment"]
                column["Tasks"] = status["task"]
                column["Total Experiments"] = status["experiments"]
                total_exp_num += status["experiments"]
                column["Completed Experiments"] = status["completed"]
                complete_exp_num += status["completed"]
                column["Submitted Trials"] = status["trials"]
                total_trials += status["trials"]
                benchmark_status.append(column)

        if not silent:
            print(
                "Benchmark name: {}, Experiments: {}/{}, Submitted trials: {}".format(
                    self.name, complete_exp_num, total_exp_num, total_trials
                )
            )
            self._pretty_table(benchmark_status)

        return benchmark_status

    def analysis(self):
        """Return all the assessment figures"""
        figures = []
        for study in self.studies:
            figure = study.analysis()
            figures.append(figure)
        return figures

    def experiments(self, silent=True):
        """Return all the experiments submitted in benchmark"""
        experiment_table = []
        for study in self.studies:
            for exp in study.experiments():
                exp_column = dict()
                stats = exp.stats
                exp_column["Algorithm"] = list(exp.configuration["algorithms"].keys())[
                    0
                ]
                exp_column["Experiment Name"] = exp.name
                exp_column["Number Trial"] = len(exp.fetch_trials())
                exp_column["Best Evaluation"] = stats["best_evaluation"]
                experiment_table.append(exp_column)

        if not silent:
            print("Total Experiments: {}".format(len(experiment_table)))
            self._pretty_table(experiment_table)

        return experiment_table

    def _pretty_table(self, dict_list):
        """
        Print a list of same format dict as pretty table with IPython disaply(notebook) or tablute
        :param dict_list: a list of dict where each dict has the same keys.
        """
        try:
            from IPython.display import HTML, display

            display(
                HTML(
                    tabulate(
                        dict_list,
                        headers="keys",
                        tablefmt="html",
                        stralign="center",
                        numalign="center",
                    )
                )
            )
        except ImportError:
            table = tabulate(
                dict_list,
                headers="keys",
                tablefmt="grid",
                stralign="center",
                numalign="center",
            )
            print(table)

    # pylint: disable=invalid-name
    @property
    def id(self):
        """Id of the benchmark in the database if configured.

        Value is `None` if the benchmark is not configured.
        """
        return self._id

    @property
    def configuration(self):
        """Return a copy of an `Benchmark` configuration as a dictionary."""
        config = dict()

        config["name"] = self.name
        config["algorithms"] = self.algorithms

        targets = []
        for target in self.targets:
            str_target = {}
            assessments = target["assess"]
            str_assessments = dict()
            for assessment in assessments:
                str_assessments.update(assessment.configuration)
            str_target["assess"] = str_assessments

            tasks = target["task"]
            str_tasks = dict()
            for task in tasks:
                str_tasks.update(task.configuration)
            str_target["task"] = str_tasks

        targets.append(str_target)
        config["targets"] = targets

        if self.id is not None:
            config["_id"] = self.id

        return copy.deepcopy(config)


class Study:
    """
    A study is one assessment and task combination in the `Benchmark` targets. It will
    build and run experiments for all the algorithms for that task.

    Parameters
    ----------
    benchmark: A Benchmark instance
    algorithms: list
        Algorithms used for benchmark, each algorithm can be a string or dict, with same format as `Benchmark` algorithms.
    assessment: list
        `Assessment` instance
    task: list
        `Task` instance
    """

    class _StudyAlgorithm:
        """
        Represent user input json format algorithm as a Study algorithm object for easy to use.
        Parameters
        ----------
        algorithm: one algorithm in the `Study` algorithms list.
        """

        def __init__(self, algorithm):
            parameters = None
            deterministic = False

            if isinstance(algorithm, dict):
                if len(algorithm) > 1 or algorithm.get("algorithm"):
                    deterministic = algorithm.get("deterministic", False)
                    experiment_algorithm = algorithm["algorithm"]

                    if isinstance(experiment_algorithm, dict):
                        name, parameters = list(experiment_algorithm.items())[0]
                    else:
                        name = experiment_algorithm
                else:
                    name, parameters = list(algorithm.items())[0]
            else:
                name = algorithm

            self.algo_name = name
            self.parameters = parameters
            self.deterministic = deterministic

        @property
        def name(self):
            return self.algo_name

        @property
        def experiment_algorithm(self):
            if self.parameters:
                return {self.algo_name: self.parameters}
            else:
                return self.algo_name

        @property
        def is_deterministic(self):
            return self.deterministic

    def __init__(self, benchmark, algorithms, assessment, task):
        self.algorithms = self._build_benchmark_algorithms(algorithms)
        self.assessment = assessment
        self.task = task
        self.benchmark = benchmark

        self.assess_name = type(self.assessment).__name__
        self.task_name = type(self.task).__name__
        self.experiments_info = []

    def _build_benchmark_algorithms(self, algorithms):
        benchmark_algorithms = list()
        for algorithm in algorithms:
            benchmark_algorithm = self._StudyAlgorithm(algorithm)
            benchmark_algorithms.append(benchmark_algorithm)
        return benchmark_algorithms

    def setup_experiments(self):
        """Setup experiments to run of the study"""
        max_trials = self.task.max_trials
        task_num = self.assessment.task_num
        space = self.task.get_search_space()

        for task_index in range(task_num):

            for algo_index, algorithm in enumerate(self.algorithms):

                # Run only 1 experiment for deterministic algorithm
                if algorithm.is_deterministic and task_index > 0:
                    continue

                experiment_name = (
                    self.benchmark.name
                    + "_"
                    + self.assess_name
                    + "_"
                    + self.task_name
                    + "_"
                    + str(task_index)
                    + "_"
                    + str(algo_index)
                )

                experiment = create_experiment(
                    experiment_name,
                    space=space,
                    algorithms=algorithm.experiment_algorithm,
                    max_trials=max_trials,
                    storage=self.benchmark.storage_config,
                )
                self.experiments_info.append((task_index, experiment))

    def execute(self, n_workers=1):
        """Execute all the experiments of the study"""
        max_trials = self.task.max_trials

        for _, experiment in self.experiments_info:
            # TODO: it is a blocking call
            experiment.workon(self.task, max_trials, n_workers)

    def status(self):
        """Return status of the study"""
        algorithm_tasks = {}

        for _, experiment in self.experiments_info:
            trials = experiment.fetch_trials()

            algorithm_name = list(experiment.configuration["algorithms"].keys())[0]

            if algorithm_tasks.get(algorithm_name, None) is None:
                task_state = {
                    "algorithm": algorithm_name,
                    "experiments": 0,
                    "assessment": self.assess_name,
                    "task": self.task_name,
                    "completed": 0,
                    "trials": 0,
                }
            else:
                task_state = algorithm_tasks[algorithm_name]

            task_state["experiments"] += 1
            task_state["trials"] += len(trials)
            if experiment.is_done:
                task_state["completed"] += 1

            algorithm_tasks[algorithm_name] = task_state

        return list(algorithm_tasks.values())

    def analysis(self):
        """Return assessment figure"""
        return self.assessment.analysis(self.task_name, self.experiments_info)

    def experiments(self):
        """Return all the experiments of the study"""
        exps = []
        for _, experiment in self.experiments_info:
            exps.append(experiment)
        return exps

    def __repr__(self):
        """Represent the object as a string."""
        algorithms_list = [algorithm.name for algorithm in self.algorithms]

        return "Study(assessment=%s, task=%s, algorithms=[%s])" % (
            self.assess_name,
            self.task_name,
            ",".join(algorithms_list),
        )



# -*- coding: utf-8 -*-
"""
Provides HPO analysis tools
===========================
"""

from orion.analysis.base import average, ranking
from orion.analysis.lpi_utils import lpi
from orion.analysis.partial_dependency_utils import partial_dependency
from orion.analysis.regret_utils import regret

__all__ = ["average", "ranking", "lpi", "partial_dependency", "regret"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.database.pickleddb`."""
import functools
from datetime import datetime
from timeit import timeit

import pymongo
import pytest
from pymongo import MongoClient
from test_database import (
    clean_db,
    db_test_data,
    drop_collections,
    get_db,
    insert_collections,
    insert_test_collection,
    orion_db,
    test_collection,
)

from orion.core.io.database import Database, DatabaseError, DuplicateKeyError
from orion.core.io.database.mongodb import AUTH_FAILED_MESSAGES, MongoDB


@pytest.fixture(scope="module", autouse=True)
def db_type(pytestconfig, request):
    """Return the string identifier of a MongoDB if the --mongodb option is
    active"""
    if not pytestconfig.getoption("--mongodb"):
        pytest.skip("mongodb tests disabled")
    yield "mongodb"


@pytest.fixture()
def patch_mongo_client(monkeypatch, db_type):
    """Patch ``pymongo.MongoClient`` to force serverSelectionTimeoutMS to 1."""
    if db_type != "mongodb":
        pytest.skip("mongodb test only")

    def mock_class(*args, **kwargs):
        # 1 sec, defaults to 20 secs otherwise
        kwargs["serverSelectionTimeoutMS"] = 1.0
        # NOTE: Can't use pymongo.MongoClient otherwise there is an infinit
        # recursion; mock(mock(mock(mock(...(MongoClient)...))))
        return MongoClient(*args, **kwargs)

    monkeypatch.setattr("pymongo.MongoClient", mock_class)


# TESTS SET


@pytest.mark.usefixtures("clean_db")
@pytest.mark.drop_collections(["new_collection"])
class TestEnsureIndex(object):
    """Calls to :meth:`orion.core.io.database.AbstractDB.ensure_index`."""

    def test_unique_index(self, orion_db):
        """Index should be set as unique in mongo database's index information."""
        assert (
            "name_1_metadata.user_1"
            not in get_db(orion_db)["experiments"].index_information()
        )
        orion_db.ensure_index(
            "experiments",
            [("name", Database.ASCENDING), ("metadata.user", Database.ASCENDING)],
            unique=True,
        )
        index_information = get_db(orion_db)["experiments"].index_information()
        assert "name_1_metadata.user_1" in index_information
        assert index_information["name_1_metadata.user_1"]["unique"]


@pytest.mark.usefixtures("clean_db")
@insert_test_collection
class TestDropIndex(object):
    """Calls :meth:`orion.core.io.database.AbstractDB.drop_index`."""

    @pytest.mark.parametrize(
        "db_test_data",
        [
            {
                "mongodb": (
                    [("name", Database.ASCENDING)],
                    [("name", Database.DESCENDING)],
                    "name_1",
                    "name_-1",
                    {"_id_": True, "name_1": False, "name_-1": False},
                    {"_id_": True, "name_-1": False},
                ),
            },
            {
                "mongodb": (
                    [("name", Database.ASCENDING), ("version", Database.DESCENDING)],
                    [("name", Database.DESCENDING), ("version", Database.ASCENDING)],
                    "name_1_version_-1",
                    "name_-1_version_1",
                    {
                        "_id_": True,
                        "name_1_version_-1": False,
                        "name_-1_version_1": False,
                    },
                    {"_id_": True, "name_-1_version_1": False},
                ),
            },
        ],
        indirect=True,
    )
    def test_drop_ordered_index(self, orion_db, db_test_data):
        """Test with single and compound indexes."""
        (
            keys_1,
            keys_2,
            stored_keys_1,
            stored_keys_2,
            index_information_initial,
            index_information_wo_1,
        ) = db_test_data
        orion_db.ensure_index("experiments", keys_1)
        orion_db.ensure_index("experiments", keys_2)
        assert orion_db.index_information("experiments") == index_information_initial
        orion_db.drop_index("experiments", stored_keys_1)
        assert orion_db.index_information("experiments") == index_information_wo_1
        with pytest.raises(DatabaseError) as exc:
            orion_db.drop_index("experiments", stored_keys_1)
        assert "index not found with name" in str(exc.value)
        orion_db.drop_index("experiments", stored_keys_2)
        assert orion_db.index_information("experiments") == {"_id_": True}

    def test_drop_unique_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index("hello", [("bonjour", Database.DESCENDING)], unique=True)
        index_info = orion_db.index_information("hello")
        assert index_info == {"_id_": True, "bonjour_-1": True}
        orion_db.drop_index("hello", "bonjour_-1")
        index_info = orion_db.index_information("hello")
        assert index_info == {"_id_": True}


# MONGODB ONLY TESTS


@pytest.mark.usefixtures("null_db_instances")
class TestConnection(object):
    """Create a :class:`orion.core.io.database.mongodb.MongoDB`, check connection cases."""

    @pytest.mark.usefixtures("patch_mongo_client")
    def test_bad_connection(self, monkeypatch):
        """Raise when connection cannot be achieved."""
        monkeypatch.setattr(
            MongoDB, "initiate_connection", MongoDB.initiate_connection.__wrapped__
        )
        with pytest.raises(pymongo.errors.ConnectionFailure) as exc_info:
            MongoDB(
                host="asdfada",
                port=123,
                name="orion",
                username="uasdfaf",
                password="paasdfss",
            )

        monkeypatch.undo()

        # Verify that the wrapper converts it properly to DatabaseError
        with pytest.raises(DatabaseError) as exc_info:
            MongoDB(
                host="asdfada",
                port=123,
                name="orion",
                username="uasdfaf",
                password="paasdfss",
            )
        assert "Connection" in str(exc_info.value)

    def test_bad_authentication(self, monkeypatch):
        """Raise when authentication cannot be achieved."""
        monkeypatch.setattr(
            MongoDB, "initiate_connection", MongoDB.initiate_connection.__wrapped__
        )
        with pytest.raises(pymongo.errors.OperationFailure) as exc_info:
            MongoDB(name="orion_test", username="uasdfaf", password="paasdfss")
        assert any(m in str(exc_info.value) for m in AUTH_FAILED_MESSAGES)

        monkeypatch.undo()

        with pytest.raises(DatabaseError) as exc_info:
            MongoDB(name="orion_test", username="uasdfaf", password="paasdfss")
        assert "Authentication" in str(exc_info.value)

    def test_connection_with_uri(self):
        """Check the case when connecting with ready `uri`."""
        orion_db = MongoDB("mongodb://user:pass@localhost/orion_test")
        assert orion_db.host == "localhost"
        assert orion_db.port == 27017
        assert orion_db.username == "user"
        assert orion_db.password == "pass"
        assert orion_db.name == "orion_test"

    def test_overwrite_uri(self):
        """Check the case when connecting with ready `uri`."""
        orion_db = MongoDB(
            "mongodb://user:pass@localhost:27017/orion_test",
            port=1231,
            name="orion",
            username="lala",
            password="pass",
        )
        assert orion_db.host == "localhost"
        assert orion_db.port == 27017
        assert orion_db.username == "user"
        assert orion_db.password == "pass"
        assert orion_db.name == "orion_test"

    def test_overwrite_partial_uri(self, monkeypatch):
        """Check the case when connecting with partial `uri`."""
        monkeypatch.setattr(MongoDB, "initiate_connection", lambda self: None)

        orion_db = MongoDB(
            "mongodb://localhost",
            port=1231,
            name="orion",
            username="lala",
            password="none",
        )
        orion_db._sanitize_attrs()
        assert orion_db.host == "localhost"
        assert orion_db.port == 1231
        assert orion_db.username == "lala"
        assert orion_db.password == "none"
        assert orion_db.name == "orion"

    def test_singleton(self):
        """Test that MongoDB class is a singleton."""
        orion_db = MongoDB(
            "mongodb://localhost",
            port=27017,
            name="orion_test",
            username="user",
            password="pass",
        )
        # reinit connection does not change anything
        orion_db.initiate_connection()
        orion_db.close_connection()
        assert MongoDB() is orion_db

    def test_change_server_timeout(self):
        """Test that the server timeout is correctly changed."""
        assert (
            timeit(
                lambda: MongoDB(
                    username="user",
                    password="pass",
                    name="orion_test",
                    serverSelectionTimeoutMS=1000,
                ),
                number=1,
            )
            <= 2
        )


@pytest.mark.usefixtures("clean_db")
@insert_test_collection
class TestExceptionWrapper(object):
    """Call to methods wrapped with `mongodb_exception_wrapper()`."""

    def test_duplicate_key_error(self, monkeypatch, orion_db, test_collection):
        """Should raise generic DuplicateKeyError."""
        # Add unique indexes to force trigger of DuplicateKeyError on write()
        orion_db.ensure_index(
            "test_collection",
            [("field0", Database.ASCENDING), ("datetime", Database.ASCENDING)],
            unique=True,
        )

        config_to_add = test_collection[0]
        config_to_add.pop("_id")

        query = {"_id": test_collection[1]["_id"]}

        # Make sure it raises pymongo.errors.DuplicateKeyError when there is no
        # wrapper
        monkeypatch.setattr(
            orion_db,
            "read_and_write",
            functools.partial(orion_db.read_and_write.__wrapped__, orion_db),
        )
        with pytest.raises(pymongo.errors.DuplicateKeyError) as exc_info:
            orion_db.read_and_write("test_collection", query, config_to_add)

        monkeypatch.undo()

        # Verify that the wrapper converts it properly to DuplicateKeyError
        with pytest.raises(DuplicateKeyError) as exc_info:
            orion_db.read_and_write("test_collection", query, config_to_add)
        assert "duplicate key error" in str(exc_info.value)

    def test_bulk_duplicate_key_error(self, monkeypatch, orion_db, test_collection):
        """Should raise generic DuplicateKeyError."""
        # Make sure it raises pymongo.errors.BulkWriteError when there is no
        # wrapper
        monkeypatch.setattr(
            orion_db, "write", functools.partial(orion_db.write.__wrapped__, orion_db)
        )
        with pytest.raises(pymongo.errors.BulkWriteError) as exc_info:
            orion_db.write("test_collection", test_collection)

        monkeypatch.undo()

        # Verify that the wrapper converts it properly to DuplicateKeyError
        with pytest.raises(DuplicateKeyError) as exc_info:
            orion_db.write("test_collection", test_collection)
        assert "duplicate key error" in str(exc_info.value)

    def test_non_converted_errors(self, orion_db, test_collection):
        """Should raise OperationFailure.

        This is because _id inside exp_config[0][0] cannot be set. It is an
        immutable key of the collection.

        """
        config_to_add = test_collection[0]

        query = {"_id": test_collection[1]["_id"]}

        with pytest.raises(pymongo.errors.OperationFailure):
            orion_db.read_and_write("test_collection", query, config_to_add)



"""
Provides public plotting API
=============================
"""
import orion.plotting.backend_plotly as backend


def lpi(
    experiment,
    with_evc_tree=True,
    model="RandomForestRegressor",
    model_kwargs=None,
    n_points=20,
    n_runs=10,
    **kwargs,
):
    """
    Make a bar plot to visualize the local parameter importance metric.

    For more information on the metric, see original paper at
    https://ml.informatik.uni-freiburg.de/papers/18-LION12-CAVE.pdf.

    Biedenkapp, André, et al. "Cave: Configuration assessment, visualization and evaluation."
    International Conference on Learning and Intelligent Optimization. Springer, Cham, 2018.

    Parameters
    ----------
    experiment: ExperimentClient or Experiment
        The orion object containing the experiment data

    with_evc_tree: bool, optional
        Fetch all trials from the EVC tree.
        Default: True

    model: str
        Name of the regression model to use. Can be one of
        - AdaBoostRegressor
        - BaggingRegressor
        - ExtraTreesRegressor
        - GradientBoostingRegressor
        - RandomForestRegressor (Default)

        Arguments for the regressor model.
    model_kwargs: dict
        Arguments for the regressor model.
    n: int
        Number of points to compute the variances. Default is 20.
    kwargs: dict
        All other plotting keyword arguments to be passed to
        :plotly:`express.line`.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided or if regressor name is invalid.

    """
    return backend.lpi(
        experiment,
        with_evc_tree=with_evc_tree,
        model=model,
        model_kwargs=model_kwargs,
        n_points=n_points,
        n_runs=n_runs,
        **kwargs,
    )


def parallel_coordinates(experiment, with_evc_tree=True, order=None, **kwargs):
    """
    Make a Parallel Coordinates Plot to visualize the effect of the hyperparameters
    on the objective.

    Parameters
    ----------
    experiment: ExperimentClient or Experiment
        The orion object containing the experiment data

    with_evc_tree: bool, optional
        Fetch all trials from the EVC tree.
        Default: True

    order: list of str or None
        Indicates the order of columns in the parallel coordinate plot. By default
        the columns are sorted alphabetically with the exception of the first column
        which is reserved for a fidelity dimension is there is one in the search space.

    kwargs: dict
        All other plotting keyword arguments to be passed to
        :plotly:`express.line`.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided.

    """
    return backend.parallel_coordinates(
        experiment, with_evc_tree=with_evc_tree, order=order, **kwargs
    )


def partial_dependencies(
    experiment,
    with_evc_tree=True,
    params=None,
    smoothing=0.85,
    verbose_hover=True,
    n_grid_points=10,
    n_samples=50,
    colorscale="Blues",
    model="RandomForestRegressor",
    model_kwargs=None,
):
    """
    Make countour plots to visualize the search space of each combination of params.

    Parameters
    ----------
    experiment: ExperimentClient or Experiment
        The orion object containing the experiment data

    with_evc_tree: bool, optional
        Fetch all trials from the EVC tree.
        Default: True

    params: list of str, optional
        Indicates the parameters to include in the plots. All parameters are included by default.

    smoothing: float, optional
        Smoothing applied to the countor plot. 0 corresponds to no smoothing. Default is 0.85.

    verbose_hover: bool
        Indicates whether to display the hyperparameter in hover tooltips. True by default.

    colorscale: str, optional
        The colorscale used for the contour plots. Supported values depends on the backend.
        Default is 'Blues'.

    n_grid_points: int, optional
        Number of points in the grid to compute partial dependency. Default is 10.

    n_samples: int, optinal
        Number of samples to randomly generate the grid used to compute the partial dependency.
        Default is 50.

    model: str
        Name of the regression model to use. Can be one of
        - AdaBoostRegressor
        - BaggingRegressor
        - ExtraTreesRegressor
        - GradientBoostingRegressor
        - RandomForestRegressor (Default)

    model_kwargs: dict, optional
        Arguments for the regressor model.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided.

    """

    return backend.partial_dependencies(
        experiment,
        with_evc_tree=with_evc_tree,
        params=params,
        smoothing=smoothing,
        verbose_hover=verbose_hover,
        n_grid_points=n_grid_points,
        n_samples=n_samples,
        colorscale=colorscale,
        model=model,
        model_kwargs=model_kwargs,
    )


def rankings(experiments, with_evc_tree=True, order_by="suggested", **kwargs):
    """
    Make a plot to visually compare the ranking of different hyper-optimization processes.

    The x-axis contain the trials and the y-axis their respective ranking.

    4 formats are supported for the experiments:

        1. List of experiments. The names of the experiments will be used for the figure labels.
        2. Dictionary of experiments. The keys of the dictionary will be used for the figure labels.
        3. List of dictionary of experiments. The keys of the dictionary will be used for the figure \
        labels. The ranking will be averaged across the dictionaries.
        4. Dictionary of list of experiments. The keys of the dictionary will be used for the figure \
        labels. A dictionary of experiments will be build grouping the i-th experiments of each \
        list to result in a list of dictionary of experiments. Behavior will be same as format 3.

    Parameters
    ----------
    experiments: list or dict
        List or dictionary of experiments.

    with_evc_tree: bool, optional
        Fetch all trials from the EVC tree.
        Default: True

    order_by: str
        Indicates how the trials should be ordered. Acceptable options are below.
        See attributes of ``Trial`` for more details.

        * 'suggested': Sort by trial suggested time (default).
        * 'reserved': Sort by trial reserved time.
        * 'completed': Sort by trial completed time.

    kwargs: dict
        All other plotting keyword arguments to be passed to
        :plotly:`express.line`.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided or order_by is invalid.

    """
    return backend.rankings(
        experiments, with_evc_tree=with_evc_tree, order_by=order_by, **kwargs
    )


def regret(
    experiment, with_evc_tree=True, order_by="suggested", verbose_hover=True, **kwargs
):
    """
    Make a plot to visualize the performance of the hyper-optimization process.

    The x-axis contain the trials and the y-axis their respective best performance.

    Parameters
    ----------
    experiment: ExperimentClient or Experiment
        The orion object containing the experiment data

    with_evc_tree: bool, optional
        Fetch all trials from the EVC tree.
        Default: True

    order_by: str
        Indicates how the trials should be ordered. Acceptable options are below.
        See attributes of ``Trial`` for more details.

        * 'suggested': Sort by trial suggested time (default).
        * 'reserved': Sort by trial reserved time.
        * 'completed': Sort by trial completed time.

    verbose_hover: bool
        Indicates whether to display the hyperparameter in hover tooltips. True by default.

    kwargs: dict
        All other plotting keyword arguments to be passed to
        :plotly:`express.line`.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided.

    """
    return backend.regret(
        experiment,
        with_evc_tree=with_evc_tree,
        order_by=order_by,
        verbose_hover=verbose_hover,
        **kwargs,
    )


def regrets(experiments, with_evc_tree=True, order_by="suggested", **kwargs):
    """
    Make a plot to visually compare the performance of different hyper-optimization processes.

    The x-axis contain the trials and the y-axis their respective best performance.

    3 formats are supported for the experiments:

        1. List of experiments. The names of the experiments will be used for the figure labels.
        2. Dictionary of experiments. The keys of the dictionary will be used for the figure labels.
        3. Dictionary of list of experiments. The keys of the dictionary will be used for the figure \
        labels. The objective of the experiments in each list will be averaged at every time step \
        (ex: across all first trials for a given list of experiments.)

    Parameters
    ----------
    experiments: list or dict
        List or dictionary of experiments.

    with_evc_tree: bool, optional
        Fetch all trials from the EVC tree.
        Default: True

    order_by: str
        Indicates how the trials should be ordered. Acceptable options are below.
        See attributes of ``Trial`` for more details.

        * 'suggested': Sort by trial suggested time (default).
        * 'reserved': Sort by trial reserved time.
        * 'completed': Sort by trial completed time.

    kwargs: dict
        All other plotting keyword arguments to be passed to
        :plotly:`express.line`.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided or order_by is invalid.

    """
    return backend.regrets(
        experiments, with_evc_tree=with_evc_tree, order_by=order_by, **kwargs
    )


PLOT_METHODS = {
    "lpi": lpi,
    "parallel_coordinates": parallel_coordinates,
    "partial_dependencies": partial_dependencies,
    "regret": regret,
    "regrets": regrets,
    "rankings": rankings,
}


SINGLE_EXPERIMENT_PLOTS = {
    "lpi": lpi,
    "parallel_coordinates": parallel_coordinates,
    "partial_dependencies": partial_dependencies,
    "regret": regret,
}


class PlotAccessor:
    """
    Make plots of ExperimentClient.

    Parameters
    ----------
    experiment : ExperimentClient
        The object for which the method is called.

    Raises
    ------
    ValueError
        If no experiment is provided.
    """

    def __init__(self, experiment):
        if not experiment:
            raise ValueError("Parameter 'experiment' is None")
        self._experiment = experiment

    def __call__(self, **kwargs):
        """
        Make different kinds of plots of ExperimentClient.

        Parameters
        ----------
        kind : str
            The kind of plot to produce:

            - 'regret' : Regret plot (default)
        """
        kind = kwargs.pop("kind", "regret")

        if kind not in PLOT_METHODS.keys():
            raise ValueError(
                f"Plot of kind '{kind}' is not one of {list(PLOT_METHODS.keys())}"
            )

        return PLOT_METHODS[kind](self._experiment, **kwargs)

    def lpi(self, **kwargs):
        """Make a bar plot of the local parameter importance metrics."""
        __doc__ = lpi.__doc__
        return self(kind="lpi", **kwargs)

    def parallel_coordinates(self, **kwargs):
        """Make a parallel coordinates plot to visualize the performance of the
        hyper-optimization process.
        """
        __doc__ = parallel_coordinates.__doc__
        return self(kind="parallel_coordinates", **kwargs)

    def partial_dependencies(self, **kwargs):
        """Make countour plots to visualize the search space of each combination of params."""
        __doc__ = partial_dependencies.__doc__
        return self(kind="partial_dependencies", **kwargs)

    def regret(self, **kwargs):
        """Make a plot to visualize the performance of the hyper-optimization process."""
        __doc__ = regret.__doc__
        return self(kind="regret", **kwargs)



# pylint:disable=too-many-lines
"""
Create experiment from user options
===================================

Functions which build :class:`orion.core.worker.experiment.Experiment` objects based on user
configuration.

The instantiation of an :class:`orion.core.worker.experiment.Experiment` is not a trivial process
when the user request an experiment with specific options. One can easily create a new experiment
with ``Experiment('some_experiment_name')``, but the configuration of a _writable_ experiment is
less straighforward. This is because there is many sources of configuration and they have a strict
hierarchy. From the more global to the more specific, there is:

1. Global configuration:

  Defined by ``orion.core.DEF_CONFIG_FILES_PATHS``.
  Can be scattered in user file system, defaults could look like:

    - ``/some/path/to/.virtualenvs/orion/share/orion.core``
    - ``/etc/xdg/xdg-ubuntu/orion.core``
    - ``/home/${USER}/.config/orion.core``

  Note that most variables have default value even if user do not defined them in global
  configuration. These are defined in ``orion.core.__init__``.

2. Oríon specific environment variables:

   Environment variables which can override global configuration

    - Database specific:

      * ``ORION_DB_NAME``
      * ``ORION_DB_TYPE``
      * ``ORION_DB_ADDRESS``

3. Experiment configuration inside the database

  Configuration of the experiment if present in the database.
  Making this part of the configuration of the experiment makes it possible
  for the user to execute an experiment by only specifying partial configuration. The rest of the
  configuration is fetched from the database.

  For example, a user could:

    1. Rerun the same experiment

      Only providing the name is sufficient to rebuild the entire configuration of the
      experiment.

    2. Make a modification to an existing experiment

      The user can provide the name of the existing experiment and only provide the changes to
      apply on it. Here is an minimal example where we fully initialize a first experiment with a
      config file and then branch from it with minimal information.

      .. code-block:: bash

          # Initialize root experiment
          orion hunt --init-only --config previous_exeriment.yaml ./userscript -x~'uniform(0, 10)'
          # Branch a new experiment
          orion hunt -n previous_experiment ./userscript -x~'uniform(0, 100)'

4. Configuration file

  This configuration file is meant to overwrite the configuration coming from the database.
  If this configuration file was interpreted as part of the global configuration, a user could
  only modify an experiment using command line arguments.

5. Command-line arguments

  Those are the arguments provided to ``orion`` for any method (hunt, insert, etc). It includes the
  argument to ``orion`` itself as well as the user's script name and its arguments.

"""
import copy
import datetime
import getpass
import logging
import pprint
import sys

import orion.core
import orion.core.utils.backward as backward
from orion.algo.space import Space
from orion.core.evc.adapters import Adapter
from orion.core.evc.conflicts import ExperimentNameConflict, detect_conflicts
from orion.core.io import resolve_config
from orion.core.io.database import DuplicateKeyError
from orion.core.io.experiment_branch_builder import ExperimentBranchBuilder
from orion.core.io.interactive_commands.branching_prompt import BranchingPrompt
from orion.core.io.space_builder import SpaceBuilder
from orion.core.utils.exceptions import (
    BranchingEvent,
    NoConfigurationError,
    NoNameError,
    RaceCondition,
)
from orion.core.worker.experiment import Experiment
from orion.core.worker.primary_algo import PrimaryAlgo
from orion.core.worker.strategy import Strategy
from orion.storage.base import get_storage, setup_storage

log = logging.getLogger(__name__)


##
# Functions to build experiments
##


def build(name, version=None, branching=None, **config):
    """Build an experiment object

    If new, ``space`` argument must be provided, else all arguments are fetched from the database
    based on (name, version). If any argument given does not match the corresponding ones in the
    database for given (name, version), than the version is incremented and the experiment will be a
    child of the previous version.

    Parameters
    ----------
    name: str
        Name of the experiment to build
    version: int, optional
        Version to select. If None, last version will be selected. If version given is larger than
        largest version available, the largest version will be selected.
    space: dict, optional
        Optimization space of the algorithm. Should have the form ``dict(name='<prior>(args)')``.
    algorithms: str or dict, optional
        Algorithm used for optimization.
    strategy: str or dict, optional
        Parallel strategy to use to parallelize the algorithm.
    max_trials: int, optional
        Maximum number of trials before the experiment is considered done.
    max_broken: int, optional
        Number of broken trials for the experiment to be considered broken.
    storage: dict, optional
        Configuration of the storage backend.

    branching: dict, optional
        Arguments to control the branching.

        branch_from: str, optional
            Name of the experiment to branch from.
        manual_resolution: bool, optional
            Starts the prompt to resolve manually the conflicts. Defaults to False.
        non_monitored_arguments: list of str, optional
            Will ignore these arguments while looking for differences. Defaults to [].
        ignore_code_changes: bool, optional
            Will ignore code changes while looking for differences. Defaults to False.
        algorithm_change: bool, optional
            Whether to automatically solve the algorithm conflict (change of algo config).
            Defaults to True.
        orion_version_change: bool, optional
            Whether to automatically solve the orion version conflict.
            Defaults to True.
        code_change_type: str, optional
            How to resolve code change automatically. Must be one of 'noeffect', 'unsure' or
            'break'.  Defaults to 'break'.
        cli_change_type: str, optional
            How to resolve cli change automatically. Must be one of 'noeffect', 'unsure' or 'break'.
            Defaults to 'break'.
        config_change_type: str, optional
            How to resolve config change automatically. Must be one of 'noeffect', 'unsure' or
            'break'.  Defaults to 'break'.

    """
    log.debug(f"Building experiment {name} with {version}")
    log.debug("    Passed experiment config:\n%s", pprint.pformat(config))
    log.debug("    Branching config:\n%s", pprint.pformat(branching))

    name, config, branching = clean_config(name, config, branching)

    config = consolidate_config(name, version, config)

    if "space" not in config:
        raise NoConfigurationError(
            "Experiment {} does not exist in DB and space was not defined.".format(name)
        )

    if len(config["space"]) == 0:
        raise NoConfigurationError("No prior found. Please include at least one.")

    experiment = create_experiment(mode="x", **copy.deepcopy(config))
    if experiment.id is None:
        log.debug("Experiment not found in DB. Now attempting registration in DB.")
        try:
            _register_experiment(experiment)
            log.debug("Experiment successfully registered in DB.")
        except DuplicateKeyError:
            log.debug(
                "Experiment registration failed. This is likely due to a race condition. "
                "Now rolling back and re-attempting building it."
            )
            experiment = build(branching=branching, **config)

        return experiment

    log.debug(f"Experiment {config['name']}-v{config['version']} already existed.")

    conflicts = _get_conflicts(experiment, branching)
    must_branch = len(conflicts.get()) > 1 or branching.get("branch_to")
    if must_branch:
        if len(conflicts.get()) > 1:
            log.debug("Experiment must branch because of conflicts")
        else:
            assert branching.get("branch_to")
            log.debug("Experiment branching forced with ``branch_to``")
        branched_experiment = _branch_experiment(
            experiment, conflicts, version, branching
        )
        log.debug("Now attempting registration of branched experiment in DB.")
        try:
            _register_experiment(branched_experiment)
            log.debug("Branched experiment successfully registered in DB.")
        except DuplicateKeyError as e:
            log.debug(
                "Experiment registration failed. This is likely due to a race condition "
                "during branching. Now rolling back and re-attempting building "
                "the branched experiment."
            )
            raise RaceCondition("There was a race condition during branching.") from e

        return branched_experiment

    log.debug("No branching required.")

    _update_experiment(experiment)
    return experiment


def clean_config(name, config, branching):
    """Clean configuration from hidden fields (ex: ``_id``) and update branching if necessary"""
    log.debug("Cleaning config")

    config = copy.deepcopy(config)
    for key, value in list(config.items()):
        if key.startswith("_") or value is None:
            log.debug(f"Ignoring field {key}")
            config.pop(key)

    if "strategy" in config:
        config["producer"] = {"strategy": config.pop("strategy")}

    if branching is None:
        branching = {}

    if branching.get("branch_from"):
        branching.setdefault("branch_to", name)
        name = branching["branch_from"]

    log.debug("Cleaned experiment config")
    log.debug("    Experiment config:\n%s", pprint.pformat(config))
    log.debug("    Branching config:\n%s", pprint.pformat(branching))

    return name, config, branching


def consolidate_config(name, version, config):
    """Merge together given configuration with db configuration matching
    for experiment (``name``, ``version``)
    """
    db_config = fetch_config_from_db(name, version)

    # Do not merge spaces, the new definition overrides it.
    if "space" in config:
        db_config.pop("space", None)

    log.debug("Merging user and db configs:")
    log.debug("    config from user:\n%s", pprint.pformat(config))
    log.debug("    config from DB:\n%s", pprint.pformat(db_config))

    new_config = config
    config = resolve_config.merge_configs(db_config, config)

    config.setdefault("metadata", {})
    resolve_config.update_metadata(config["metadata"])

    merge_algorithm_config(config, new_config)
    merge_producer_config(config, new_config)

    config.setdefault("name", name)
    config.setdefault("version", version)

    log.debug("    Merged config:\n%s", pprint.pformat(config))

    return config


def merge_algorithm_config(config, new_config):
    """Merge given algorithm configuration with db config"""
    # TODO: Find a better solution
    if isinstance(config.get("algorithms"), dict) and len(config["algorithms"]) > 1:
        log.debug("Overriding algo config with new one.")
        log.debug("    Old config:\n%s", pprint.pformat(config["algorithms"]))
        log.debug("    New config:\n%s", pprint.pformat(new_config["algorithms"]))
        config["algorithms"] = new_config["algorithms"]


def merge_producer_config(config, new_config):
    """Merge given producer configuration with db config"""
    # TODO: Find a better solution
    if (
        isinstance(config.get("producer", {}).get("strategy"), dict)
        and len(config["producer"]["strategy"]) > 1
    ):
        log.debug("Overriding strategy config with new one.")
        log.debug("    Old config:\n%s", pprint.pformat(config["producer"]["strategy"]))
        log.debug(
            "    New config:\n%s", pprint.pformat(new_config["producer"]["strategy"])
        )

        config["producer"]["strategy"] = new_config["producer"]["strategy"]


def build_view(name, version=None):
    """Load experiment from database

    This function is deprecated and will be remove in v0.3.0. Use `load()` instead.
    """
    return load(name, version=version, mode="r")


def load(name, version=None, mode="r"):
    """Load experiment from database

    An experiment view provides all reading operations of standard experiment but prevents the
    modification of the experiment and its trials.

    Parameters
    ----------
    name: str
        Name of the experiment to build
    version: int, optional
        Version to select. If None, last version will be selected. If version given is larger than
        largest version available, the largest version will be selected.
    mode: str, optional
        The access rights of the experiment on the database.
        'r': read access only
        'w': can read and write to database
        Default is 'r'

    """
    assert mode in set("rw")

    log.debug(
        f"Loading experiment {name} (version={version}) from database in mode `{mode}`"
    )
    db_config = fetch_config_from_db(name, version)

    if not db_config:
        message = (
            "No experiment with given name '%s' and version '%s' inside database, "
            "no view can be created." % (name, version if version else "*")
        )
        raise NoConfigurationError(message)

    db_config.setdefault("version", 1)

    return create_experiment(mode=mode, **db_config)


def create_experiment(name, version, mode, space, **kwargs):
    """Instantiate the experiment and its attribute objects

    All unspecified arguments will be replaced by system's defaults (orion.core.config.*).

    Parameters
    ----------
    name: str
        Name of the experiment.
    version: int
        Version of the experiment.
    mode: str
        The access rights of the experiment on the database.
        'r': read access only
        'w': can read and write to database
        'x': can read and write to database, algo is instantiated and can execute optimization
    space: dict or Space object
        Optimization space of the algorithm. If dict, should have the form
        `dict(name='<prior>(args)')`.
    algorithms: str or dict, optional
        Algorithm used for optimization.
    strategy: str or dict, optional
        Parallel strategy to use to parallelize the algorithm.
    max_trials: int, optional
        Maximum number or trials before the experiment is considered done.
    max_broken: int, optional
        Number of broken trials for the experiment to be considered broken.
    storage: dict, optional
        Configuration of the storage backend.

    """
    experiment = Experiment(name=name, version=version, mode=mode)
    experiment._id = kwargs.get("_id", None)  # pylint:disable=protected-access
    experiment.pool_size = kwargs.get("pool_size")
    if experiment.pool_size is None:
        experiment.pool_size = orion.core.config.experiment.get(
            "pool_size", deprecated="ignore"
        )
    experiment.max_trials = kwargs.get(
        "max_trials", orion.core.config.experiment.max_trials
    )
    experiment.max_broken = kwargs.get(
        "max_broken", orion.core.config.experiment.max_broken
    )
    experiment.space = _instantiate_space(space)
    experiment.algorithms = _instantiate_algo(
        experiment.space,
        experiment.max_trials,
        kwargs.get("algorithms"),
        ignore_unavailable=mode != "x",
    )
    experiment.producer = kwargs.get("producer", {})
    experiment.producer["strategy"] = _instantiate_strategy(
        experiment.producer.get("strategy"), ignore_unavailable=mode != "x"
    )
    experiment.working_dir = kwargs.get(
        "working_dir", orion.core.config.experiment.working_dir
    )
    experiment.metadata = kwargs.get(
        "metadata", {"user": kwargs.get("user", getpass.getuser())}
    )
    experiment.refers = kwargs.get(
        "refers", {"parent_id": None, "root_id": None, "adapter": []}
    )
    experiment.refers["adapter"] = _instantiate_adapters(
        experiment.refers.get("adapter", [])
    )

    log.debug(
        "Created experiment with config:\n%s", pprint.pformat(experiment.configuration)
    )

    return experiment


def fetch_config_from_db(name, version=None):
    """Fetch configuration from database

    Parameters
    ----------
    name: str
        Name of the experiment to fetch
    version: int, optional
        Version to select. If None, last version will be selected. If version given is larger than
        largest version available, the largest version will be selected.

    """
    configs = get_storage().fetch_experiments({"name": name})

    if not configs:
        return {}

    config = _fetch_config_version(configs, version)

    if len(configs) > 1 and version is None:
        log.info(
            "Many versions for experiment %s have been found. Using latest "
            "version %s.",
            name,
            config["version"],
        )

    log.debug("Config found in DB:\n%s", pprint.pformat(config))

    backward.populate_space(config, force_update=False)
    backward.update_max_broken(config)

    return config


##
# Private helper functions to build experiments
##


def _instantiate_adapters(config):
    """Instantiate the adapter object

    Parameters
    ----------
    config: list
         List of adapter configurations to build a CompositeAdapter for the EVC.

    """
    return Adapter.build(config)


def _instantiate_space(config):
    """Instantiate the space object

    Build the Space object if argument is a dictionary, else return the Space object as is.

    Parameters
    ----------
    config: dict or Space object
        Dictionary of priors or already built Space object.

    """
    if isinstance(config, Space):
        return config

    return SpaceBuilder().build(config)


def _instantiate_algo(space, max_trials, config=None, ignore_unavailable=False):
    """Instantiate the algorithm object

    Parameters
    ----------
    config: dict, optional
        Configuration of the algorithm. If None or empty, system's defaults are used
        (orion.core.config.experiment.algorithms).
    ignore_unavailable: bool, optional
        If True and algorithm is not available (plugin not installed), return the configuration.
        Otherwise, raise Factory error from PrimaryAlgo

    """
    if not config:
        config = orion.core.config.experiment.algorithms

    try:
        algo = PrimaryAlgo(space, config)
        algo.algorithm.max_trials = max_trials
    except NotImplementedError as e:
        if not ignore_unavailable:
            raise e
        log.warning(str(e))
        log.warning("Algorithm will not be instantiated.")
        algo = config

    return algo


def _instantiate_strategy(config=None, ignore_unavailable=False):
    """Instantiate the strategy object

    Parameters
    ----------
    config: dict, optional
        Configuration of the strategy. If None of empty, system's defaults are used
        (orion.core.config.producer.strategy).
    ignore_unavailable: bool, optional
        If True and algorithm is not available (plugin not installed), return the configuration.
        Otherwise, raise Factory error from PrimaryAlgo


    """
    if not config:
        config = orion.core.config.experiment.strategy

    if isinstance(config, str):
        strategy_type = config
        config = {}
    else:
        config = copy.deepcopy(config)
        strategy_type, config = next(iter(config.items()))

    try:
        strategy = Strategy(of_type=strategy_type, **config)
    except NotImplementedError as e:
        if not ignore_unavailable:
            raise e
        log.warning(str(e))
        log.warning("Strategy will not be instantiated.")
        strategy = {strategy_type: config}

    return strategy


def _register_experiment(experiment):
    """Register a new experiment in the database"""
    experiment.metadata["datetime"] = datetime.datetime.utcnow()
    config = experiment.configuration
    # This will raise DuplicateKeyError if a concurrent experiment with
    # identical (name, metadata.user) is written first in the database.

    get_storage().create_experiment(config)

    # XXX: Reminder for future DB implementations:
    # MongoDB, updates an inserted dict with _id, so should you :P
    experiment._id = config["_id"]  # pylint:disable=protected-access

    # Update refers in db if experiment is root
    if experiment.refers.get("parent_id") is None:
        log.debug("update refers (name: %s)", experiment.name)
        experiment.refers["root_id"] = experiment.id
        get_storage().update_experiment(
            experiment, refers=experiment.configuration["refers"]
        )


def _update_experiment(experiment):
    """Update experiment configuration in database"""
    log.debug("Updating experiment (name: %s)", experiment.name)
    config = experiment.configuration

    # TODO: Remove since this should not occur anymore without metadata.user in the indices?
    # Writing the final config to an already existing experiment raises
    # a DuplicatKeyError because of the embedding id `metadata.user`.
    # To avoid this `final_config["name"]` is popped out before
    # `db.write()`, thus seamingly breaking  the compound index
    # `(name, metadata.user)`
    config.pop("name")

    get_storage().update_experiment(experiment, **config)

    log.debug("Experiment configuration successfully updated in DB.")


def _branch_experiment(experiment, conflicts, version, branching_arguments):
    """Create a new branch experiment with adapters for the given conflicts"""
    experiment_brancher = ExperimentBranchBuilder(conflicts, **branching_arguments)

    needs_manual_resolution = (
        not experiment_brancher.is_resolved or experiment_brancher.manual_resolution
    )

    if not experiment_brancher.is_resolved:
        name_conflict = conflicts.get([ExperimentNameConflict])[0]
        if not name_conflict.is_resolved and not version:
            log.debug(
                "A race condition likely occured during conflicts resolutions. "
                "Now rolling back and attempting re-building the branched experiment."
            )
            raise RaceCondition(
                "There was likely a race condition during version increment."
            )

    if needs_manual_resolution:
        log.debug("Some conflicts cannot be solved automatically.")

        # TODO: This should only be possible when using cmdline API
        branching_prompt = BranchingPrompt(experiment_brancher)

        if not sys.__stdin__.isatty():
            log.debug("No interactive prompt available to manually resolve conflicts.")
            raise BranchingEvent(branching_prompt.get_status())

        branching_prompt.cmdloop()

        if branching_prompt.abort or not experiment_brancher.is_resolved:
            sys.exit()

    log.debug("Creating new branched configuration")
    config = experiment_brancher.conflicting_config
    config["refers"]["adapter"] = experiment_brancher.create_adapters().configuration
    config["refers"]["parent_id"] = experiment.id

    config.pop("_id")

    return create_experiment(mode="x", **config)


def _get_conflicts(experiment, branching):
    """Get conflicts between current experiment and corresponding configuration in database"""
    log.debug("Looking for conflicts in new configuration.")
    db_experiment = load(experiment.name, experiment.version, mode="r")
    conflicts = detect_conflicts(
        db_experiment.configuration, experiment.configuration, branching
    )

    log.debug(f"{len(conflicts.get())} conflicts detected:\n {conflicts.get()}")

    # elif must_branch and not enable_branching:
    #     raise ValueError("Configuration is different and generate a branching event")

    return conflicts


def _fetch_config_version(configs, version=None):
    """Fetch the experiment configuration corresponding to the given version

    Parameters
    ----------
    configs: list
        List of configurations fetched from storoge.
    version: int, optional
        Version to select. If None, last version will be selected. If version given is larger than
        largest version available, the largest version will be selected.

    """
    max_version = max(configs, key=lambda exp: exp.get("version", 1)).get("version", 1)

    if version is None:
        version = max_version
    else:
        version = version

    if version > max_version:
        log.warning(
            "Version %s was specified but most recent version is only %s. " "Using %s.",
            version,
            max_version,
            max_version,
        )

    version = min(version, max_version)

    configs = filter(lambda exp: exp.get("version", 1) == version, configs)

    return next(iter(configs))


###
# Functions for commandline API
###


def build_from_args(cmdargs):
    """Build an experiment based on commandline arguments.

    Options provided in commandline and configuration file (--config) will overwrite system's
    default values and configuration from database if experiment already exits.
    Commandline arguments have precedence over configuration file options.

    .. seealso::

        :func:`orion.core.io.experiment_builder.build` for more information on experiment creation.

    """
    cmd_config = get_cmd_config(cmdargs)

    if "name" not in cmd_config:
        raise NoNameError()

    setup_storage(cmd_config["storage"], debug=cmd_config.get("debug"))

    return build(**cmd_config)


def get_from_args(cmdargs, mode="r"):
    """Build an experiment view based on commandline arguments

    .. seealso::

        :func:`orion.core.io.experiment_builder.load` for more information on creation of read-only
        experiments.

    """
    cmd_config = get_cmd_config(cmdargs)

    if "name" not in cmd_config:
        raise NoNameError()

    setup_storage(cmd_config["storage"], debug=cmd_config.get("debug"))

    name = cmd_config.get("name")
    version = cmd_config.get("version")

    return load(name, version, mode=mode)


def get_cmd_config(cmdargs):
    """Fetch configuration defined by commandline and local configuration file.

    Arguments of commandline have priority over options in configuration file.
    """
    cmdargs = resolve_config.fetch_config_from_cmdargs(cmdargs)
    cmd_config = resolve_config.fetch_config(cmdargs)
    cmd_config = resolve_config.merge_configs(cmd_config, cmdargs)

    cmd_config.update(cmd_config.pop("experiment", {}))
    cmd_config["user_script_config"] = cmd_config.get("worker", {}).get(
        "user_script_config", None
    )

    cmd_config["branching"] = cmd_config.pop("evc", {})

    # TODO: We should move branching specific stuff below in a centralized place for EVC stuff.
    if (
        cmd_config["branching"].get("auto_resolution", False)
        and cmdargs.get("manual_resolution", None) is None
    ):
        cmd_config["branching"]["manual_resolution"] = False

    non_monitored_arguments = cmdargs.get("non_monitored_arguments")
    if non_monitored_arguments:
        cmd_config["branching"][
            "non_monitored_arguments"
        ] = non_monitored_arguments.split(":")

    # TODO: user_args won't be defined if reading from DB only (`orion hunt -n <exp> ` alone)
    metadata = resolve_config.fetch_metadata(
        cmd_config.get("user"),
        cmd_config.get("user_args"),
        cmd_config.get("user_script_config"),
    )
    cmd_config["metadata"] = metadata
    cmd_config.pop("config", None)

    cmd_config["space"] = cmd_config["metadata"].get("priors", None)

    backward.update_db_config(cmd_config)

    return cmd_config



"""
.. _executor_backends:

***************
Scaling workers
***************

This tutorial shows simple examples to use the different backends of Oríon to scale execution to
multiple workers in parallel.
The parallelization of Oríon workers is explained in more details in section :ref:`parallelism`.

We will start with a basic example using scikit-learn.
"""
import joblib
import numpy
from sklearn import datasets
from sklearn.model_selection import cross_validate
from sklearn.svm import SVC


def main(C, gamma, tol, class_weight, joblib_backend="loky"):

    digits = datasets.load_digits()

    X = digits.data
    y = digits.target

    model = SVC(kernel="rbf", C=C, gamma=gamma, tol=tol, class_weight=class_weight)

    # Single metric evaluation using cross_validate
    with joblib.parallel_backend(joblib_backend):
        cv_results = cross_validate(model, X, y, cv=5)

    accuracy = numpy.mean(cv_results["test_score"])
    error_rate = 1 - accuracy

    return [{"name": "test_error_rate", "type": "objective", "value": error_rate}]


#%%
# We create a ``main()`` function that takes as arguments the hyperparameters for an SVM for
# classification. We add the argument ``joblib_backend`` so that we can control which backend
# is used to paralellize the cross-validation. By default we will use ``loky`` for multi-processing.
# We load the dataset ``digits`` and divide it in features ``X`` and targets ``y``.
# We then create the model with the given hyperparameter values.
# We use ``joblib.parallel_backend`` to execute the cross-validation on 5 folds in parallel
# for more effiency. Finaly we compute the average accuracy, and convert it to test error
# rate since Oríon is minimizing the objective and we would not want to minimize the accuracy.
# The results is returned in the format required by Oríon.
#
# We will now create an experiment with Oríon to optimize this function.

# Specify the database where the experiments are stored. We use a local PickleDB here.
storage = {
    "type": "legacy",
    "database": {
        "type": "pickleddb",
        "host": "./db.pkl",
    },
}

# Specify optimization space for the SVM
space = {
    "C": "loguniform(1e-6, 1e6)",
    "gamma": "loguniform(1e-8, 1e8)",
    "tol": "loguniform(1e-4, 1e-1)",
    "class_weight": "choices([None, 'balanced'])",
}


#%%
# Joblib
# ------
#
# `Joblib`_ is a lightweight library for task parallel execution in Python. It is the default
# backend used by Oríon to spawn multiple workers.
#
# We first build the experiment and limit it to 200 trials.

from orion.client import build_experiment

experiment = build_experiment(
    name="joblib_example",
    max_trials=200,
    space=space,
    storage=storage,
)

#%%
# Since joblib is the default backend, we do not need to do anything special to use it.
# We can simply call
# :meth:`ExperimentClient.workon() <orion.client.experiment.ExperimentClient.workon>`
# and specify the number of workers that we want.


experiment.workon(main, n_workers=4)

#%%
# It is as simple as this.
#
# The experiment backend is by default the one configured in the global configuration
# (:ref:`config_worker_executor` and :ref:`config_worker_executor_configuration`).
# If you want to use a different backend while executing
# :meth:`ExperimentClient.workon() <orion.client.experiment.ExperimentClient.workon>`, you can use
# :meth:`ExperimentClient.tmp_executor() <orion.client.experiment.ExperimentClient.tmp_executor>`
# like the following.


with experiment.tmp_executor("joblib", n_workers=10):
    experiment.workon(main, n_workers=2)

#%%
# Note that you must specify ``n_workers`` for both the backend and for
# :meth:`ExperimentClient.workon() <orion.client.experiment.ExperimentClient.workon>`.
# For the backend, it
# refers to the number of workers you would like to have running in parallel.
# For :meth:`ExperimentClient.workon() <orion.client.experiment.ExperimentClient.workon>`,
# it is the number of
# Orion workers to run in parallel. You may want to have more backend workers than
# Oríon workers if for instance your task is also parallelizing tasks. You can
# see such an example here, because we are parallelizing the cross-validation inside
# the function Oríon is optimizing. Each worker will create 5 tasks that can be run in parallel.
#
# .. _Joblib: https://joblib.readthedocs.io/en/latest/
#
# .. _parallel_backend(): https://joblib.readthedocs.io/en/latest/parallel.html#joblib.parallel_backend
#
# Dask
# ----
#
# Using Dask is similar to joblib. Unless you have configured
# Oríon to use Dask by default
# (:ref:`config_worker_executor` and :ref:`config_worker_executor_configuration`),
# you will want to use
# :meth:`ExperimentClient.tmp_executor() <orion.client.experiment.ExperimentClient.tmp_executor>`
# to change the backend. Make sure to only run Dask with if ``__name__ == "__main__"``
# otherwise you will run into a RuntimeError
# (`see here <https://github.com/dask/distributed/issues/2520>`_).

experiment = build_experiment(
    name="dask_example",
    max_trials=200,
    space=space,
    storage=storage,
)

if __name__ == "__main__":
    with experiment.tmp_executor("dask", n_workers=10):
        experiment.workon(main, n_workers=2, joblib_backend="dask")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module to check if the DB works
===============================

Runs multiple checks to see if the database was correctly setup.

"""
import argparse
import logging

from orion.core.cli.db.test import main

log = logging.getLogger(__name__)
DESCRIPTION = "(DEPRECATED) Use command `orion db test` instead"


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    test_db_parser = parser.add_parser(
        "test-db", help=DESCRIPTION, description=DESCRIPTION
    )

    test_db_parser.add_argument(
        "-c",
        "--config",
        type=argparse.FileType("r"),
        metavar="path-to-config",
        help="user provided " "orion configuration file",
    )

    test_db_parser.set_defaults(func=wrap_main)

    return test_db_parser


def wrap_main(args):
    """Run through all checks for database."""
    log.warning(
        "Command `orion test-db` is deprecated and will be removed in v0.2.0. Use "
        "`orion db test` instead."
    )

    main(args)



import argparse
import glob
import os
import shutil
import subprocess
import sys

from orion.client import get_experiment
from orion.core.cli.db.rm import process_exp_rm
from orion.core.utils.singleton import update_singletons
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage, setup_storage

ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
DOC_SRC_DIR = os.path.join(ROOT_DIR, "..", "src")
os.chdir(DOC_SRC_DIR)

EXAMPLE_DIR = os.path.abspath("../../examples")

MAIN_DB_HOST = f"{EXAMPLE_DIR}/db.pkl"
BASE_DB_HOST = f"{EXAMPLE_DIR}/base_db.pkl"
TMP_DB_HOST = f"{EXAMPLE_DIR}/tmp.pkl"

custom_plots = {
    "hyperband-cifar10": {
        "name": "params",
        "kwargs": {
            "kind": "partial_dependencies",
            "params": ["gamma", "learning_rate"],
        },
    },
    "dask": {
        "name": "params",
        "kwargs": {
            "kind": "partial_dependencies",
            "params": ["C", "gamma"],
        },
    },
}

# CP base db to database.pkl (overwrite database.pkl)

CODE_PATH = f"{EXAMPLE_DIR}/tutorials/{{example}}.py"

paths = glob.glob(CODE_PATH.format(example="code_*"))
names = sorted(os.path.splitext(os.path.basename(path))[0] for path in paths)


def execute(example):
    command = f"python {EXAMPLE_DIR}/tutorials/{example}.py"
    os.chdir(f"{EXAMPLE_DIR}")
    print("executing", command)
    process = subprocess.Popen(command.split(" "))
    return_code = process.wait()
    tmp_files = "tmp_" + "_".join(example.split("_")[2:])
    print("removing tmp files", tmp_files)
    shutil.rmtree(tmp_files, ignore_errors=True)
    os.chdir(DOC_SRC_DIR)
    print("done")
    return return_code


def prepare_dbs():

    if os.path.exists(MAIN_DB_HOST):
        print("Removing", MAIN_DB_HOST)
        os.remove(MAIN_DB_HOST)

    print("Copying", BASE_DB_HOST, "->", TMP_DB_HOST)
    shutil.copy(BASE_DB_HOST, TMP_DB_HOST)


def setup_tmp_storage(host):
    # Clear singletons
    update_singletons()

    setup_storage(
        storage={
            "type": "legacy",
            "database": {
                "type": "pickleddb",
                "host": host,
            },
        }
    )

    return get_storage()


def load_data(host):
    print("Loading data from", host)
    storage = setup_tmp_storage(host)
    experiment_names = set()
    data = {"experiments": {}, "trials": {}}
    for experiment in storage.fetch_experiments({}):
        data["experiments"][experiment["_id"]] = experiment
        data["trials"][experiment["_id"]] = storage.fetch_trials(uid=experiment["_id"])
        experiment_names.add((experiment["name"], experiment["version"]))

    return experiment_names, data


def copy_data(data, host=TMP_DB_HOST):
    print("Copying data to", host)
    storage = setup_tmp_storage(host)
    for exp_id, experiment in data["experiments"].items():
        del experiment["_id"]
        storage.create_experiment(experiment)
        assert exp_id != experiment["_id"]
        trials = []
        for trial in data["trials"][exp_id]:
            trial.experiment = experiment["_id"]
            trials.append(trial.to_dict())
        storage._db.write("trials", trials)


def plot_exps(experiment_names, host=TMP_DB_HOST):
    print("Plotting experiments from", host)
    storage = setup_tmp_storage(host)
    # Plot exps
    for experiment_name, version in experiment_names:
        print(f"   {experiment_name}-v{version}")
        experiment = get_experiment(experiment_name, version=version)
        for plot in ["regret", "lpi", "partial_dependencies", "parallel_coordinates"]:
            experiment.plot(kind=plot).write_html(
                f"_static/{experiment.name}_{plot}.html"
            )

        if experiment_name in custom_plots:
            custom_plot = custom_plots[experiment_name]
            kwargs = custom_plot["kwargs"]
            name = (
                f"_static/{experiment.name}_{kwargs['kind']}_{custom_plot['name']}.html"
            )
            experiment.plot(**kwargs).write_html(name)


def main(argv=None):

    parser = argparse.ArgumentParser()
    parser.add_argument("examples", nargs="*", default=[], choices=names + [[]])
    options = parser.parse_args(argv)

    if options.examples:
        run = {name for name in options.examples if name in names}
    else:
        run = set()

    prepare_dbs()

    for example in names:

        example_db_host = f"{EXAMPLE_DIR}/{example}_db.pkl"

        if example in run:
            execute(example)
            print("Moving", MAIN_DB_HOST, "->", example_db_host)
            os.rename(MAIN_DB_HOST, example_db_host)

        experiment_names, data = load_data(example_db_host)
        copy_data(data, TMP_DB_HOST)

        if example in run:
            plot_exps(experiment_names, TMP_DB_HOST)

    print("Moving", TMP_DB_HOST, "->", MAIN_DB_HOST)
    os.rename(TMP_DB_HOST, MAIN_DB_HOST)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.client`."""
import copy
import json
from importlib import reload

import pytest

import orion.client
import orion.client.cli as cli
import orion.core
from orion.client import get_experiment
from orion.client.experiment import ExperimentClient
from orion.core.io.database.ephemeraldb import EphemeralDB
from orion.core.io.database.pickleddb import PickledDB
from orion.core.utils.exceptions import (
    BranchingEvent,
    NoConfigurationError,
    RaceCondition,
    UnsupportedOperation,
)
from orion.core.utils.singleton import SingletonNotInstantiatedError, update_singletons
from orion.storage.base import get_storage
from orion.storage.legacy import Legacy
from orion.testing import OrionState

create_experiment = orion.client.create_experiment
workon = orion.client.workon


config = dict(
    name="supernaekei",
    space={"x": "uniform(0, 200)"},
    metadata={
        "user": "tsirif",
        "orion_version": "XYZ",
        "VCS": {
            "type": "git",
            "is_dirty": False,
            "HEAD_sha": "test",
            "active_branch": None,
            "diff_sha": "diff",
        },
    },
    version=1,
    pool_size=1,
    max_trials=10,
    max_broken=5,
    working_dir="",
    algorithms={"random": {"seed": 1}},
    producer={"strategy": "NoParallelStrategy"},
    refers=dict(root_id="supernaekei", parent_id=None, adapter=[]),
)


@pytest.fixture()
def user_config():
    """Curate config as a user would provide it"""
    user_config = copy.deepcopy(config)
    user_config.pop("metadata")
    user_config.pop("version")
    user_config["strategy"] = user_config.pop("producer")["strategy"]
    user_config.pop("refers")
    user_config.pop("pool_size")
    return user_config


@pytest.fixture()
def data():
    """Return serializable data."""
    return "this is datum"


class TestReportResults(object):
    """Check functionality and edge cases of `report_results` helper interface."""

    def test_with_no_env(self, monkeypatch, capsys, data):
        """Test without having set the appropriate environmental variable.

        Then: It should print `data` parameter instead to stdout.
        """
        monkeypatch.delenv("ORION_RESULTS_PATH", raising=False)
        reloaded_client = reload(cli)

        assert reloaded_client.IS_ORION_ON is False
        assert reloaded_client.RESULTS_FILENAME is None
        assert reloaded_client._HAS_REPORTED_RESULTS is False

        reloaded_client.report_results(data)
        out, err = capsys.readouterr()
        assert reloaded_client._HAS_REPORTED_RESULTS is True
        assert out == data + "\n"
        assert err == ""

    def test_with_correct_env(self, monkeypatch, capsys, tmpdir, data):
        """Check that a file with correct data will be written to an existing
        file in a legit path.
        """
        path = str(tmpdir.join("naedw.txt"))
        with open(path, mode="w"):
            pass
        monkeypatch.setenv("ORION_RESULTS_PATH", path)
        reloaded_client = reload(cli)

        assert reloaded_client.IS_ORION_ON is True
        assert reloaded_client.RESULTS_FILENAME == path
        assert reloaded_client._HAS_REPORTED_RESULTS is False

        reloaded_client.report_results(data)
        out, err = capsys.readouterr()
        assert reloaded_client._HAS_REPORTED_RESULTS is True
        assert out == ""
        assert err == ""

        with open(path, mode="r") as results_file:
            res = json.load(results_file)
        assert res == data

    def test_with_env_set_but_no_file_exists(self, monkeypatch, tmpdir, data):
        """Check that a Warning will be raised at import time,
        if environmental is set but does not correspond to an existing file.
        """
        path = str(tmpdir.join("naedw.txt"))
        monkeypatch.setenv("ORION_RESULTS_PATH", path)

        with pytest.raises(RuntimeWarning) as exc:
            reload(cli)

        assert "existing file" in str(exc.value)

    def test_call_interface_twice(self, monkeypatch, data):
        """Check that a Warning will be raised at call time,
        if function has already been called once.
        """
        monkeypatch.delenv("ORION_RESULTS_PATH", raising=False)
        reloaded_client = reload(cli)

        reloaded_client.report_results(data)
        with pytest.raises(RuntimeWarning) as exc:
            reloaded_client.report_results(data)

        assert "already reported" in str(exc.value)
        assert reloaded_client.IS_ORION_ON is False
        assert reloaded_client.RESULTS_FILENAME is None
        assert reloaded_client._HAS_REPORTED_RESULTS is True


@pytest.mark.usefixtures("version_XYZ")
class TestCreateExperiment:
    """Test creation of experiment with `client.create_experiment()`"""

    @pytest.mark.usefixtures("setup_pickleddb_database")
    def test_create_experiment_no_storage(self, monkeypatch):
        """Test creation if storage is not configured"""
        name = "oopsie_forgot_a_storage"
        host = orion.core.config.storage.database.host

        with OrionState(storage=orion.core.config.storage.to_dict()) as cfg:
            # Reset the Storage and drop instances so that get_storage() would fail.
            cfg.cleanup()
            cfg.singletons = update_singletons()

            # Make sure storage must be instantiated during `create_experiment()`
            with pytest.raises(SingletonNotInstantiatedError):
                get_storage()

            experiment = create_experiment(name=name, space={"x": "uniform(0, 10)"})

            assert isinstance(experiment._experiment._storage, Legacy)
            assert isinstance(experiment._experiment._storage._db, PickledDB)
            assert experiment._experiment._storage._db.host == host

    def test_create_experiment_new_no_space(self):
        """Test that new experiment needs space"""
        with OrionState():
            name = "oopsie_forgot_a_space"
            with pytest.raises(NoConfigurationError) as exc:
                create_experiment(name=name)

            assert "Experiment {} does not exist in DB".format(name) in str(exc.value)

    def test_create_experiment_bad_storage(self):
        """Test error message if storage is not configured properly"""
        name = "oopsie_bad_storage"
        # Make sure there is no existing storage singleton
        update_singletons()

        with pytest.raises(NotImplementedError) as exc:
            create_experiment(
                name=name,
                storage={"type": "legacy", "database": {"type": "idontexist"}},
            )

        assert (
            "Could not find implementation of AbstractDB, type = 'idontexist'"
            in str(exc.value)
        )

    def test_create_experiment_new_default(self):
        """Test creating a new experiment with all defaults"""
        name = "all_default"
        space = {"x": "uniform(0, 10)"}
        with OrionState():
            experiment = create_experiment(name="all_default", space=space)

            assert experiment.name == name
            assert experiment.space.configuration == space

            assert experiment.max_trials == orion.core.config.experiment.max_trials
            assert experiment.max_broken == orion.core.config.experiment.max_broken
            assert experiment.working_dir == orion.core.config.experiment.working_dir
            assert experiment.algorithms.configuration == {"random": {"seed": None}}
            assert experiment.configuration["producer"] == {
                "strategy": {"MaxParallelStrategy": {"default_result": float("inf")}}
            }

    def test_create_experiment_new_full_config(self, user_config):
        """Test creating a new experiment by specifying all attributes."""
        with OrionState():
            experiment = create_experiment(**user_config)

            exp_config = experiment.configuration

            assert exp_config["space"] == config["space"]
            assert exp_config["max_trials"] == config["max_trials"]
            assert exp_config["max_broken"] == config["max_broken"]
            assert exp_config["working_dir"] == config["working_dir"]
            assert exp_config["algorithms"] == config["algorithms"]
            assert exp_config["producer"] == config["producer"]

    def test_create_experiment_hit_no_branch(self, user_config):
        """Test creating an existing experiment by specifying all identical attributes."""
        with OrionState(experiments=[config]):
            experiment = create_experiment(**user_config)

            exp_config = experiment.configuration

            assert experiment.name == config["name"]
            assert experiment.version == 1
            assert exp_config["space"] == config["space"]
            assert exp_config["max_trials"] == config["max_trials"]
            assert exp_config["max_broken"] == config["max_broken"]
            assert exp_config["working_dir"] == config["working_dir"]
            assert exp_config["algorithms"] == config["algorithms"]
            assert exp_config["producer"] == config["producer"]

    def test_create_experiment_hit_no_config(self):
        """Test creating an existing experiment by specifying the name only."""
        with OrionState(experiments=[config]):
            experiment = create_experiment(config["name"])

            assert experiment.name == config["name"]
            assert experiment.version == 1
            assert experiment.space.configuration == config["space"]
            assert experiment.algorithms.configuration == config["algorithms"]
            assert experiment.max_trials == config["max_trials"]
            assert experiment.max_broken == config["max_broken"]
            assert experiment.working_dir == config["working_dir"]
            assert (
                experiment.producer["strategy"].configuration
                == config["producer"]["strategy"]
            )

    def test_create_experiment_hit_branch(self):
        """Test creating a differing experiment that cause branching."""
        with OrionState(experiments=[config]):
            experiment = create_experiment(
                config["name"], space={"y": "uniform(0, 10)"}
            )

            assert experiment.name == config["name"]
            assert experiment.version == 2

            assert experiment.algorithms.configuration == config["algorithms"]
            assert experiment.max_trials == config["max_trials"]
            assert experiment.max_broken == config["max_broken"]
            assert experiment.working_dir == config["working_dir"]
            assert (
                experiment.producer["strategy"].configuration
                == config["producer"]["strategy"]
            )

    def test_create_experiment_race_condition(self, monkeypatch):
        """Test that a single race condition is handled seemlessly

        RaceCondition during registration is already handled by `build()`, therefore we will only
        test for race conditions during version update.
        """
        with OrionState(experiments=[config]):
            parent = create_experiment(config["name"])
            child = create_experiment(config["name"], space={"y": "uniform(0, 10)"})

            def insert_race_condition(self, query):
                is_auto_version_query = query == {
                    "name": config["name"],
                    "refers.parent_id": parent.id,
                }
                if is_auto_version_query:
                    data = [child.configuration]
                # First time the query returns no other child
                elif insert_race_condition.count < 1:
                    data = [parent.configuration]
                else:
                    data = [parent.configuration, child.configuration]

                insert_race_condition.count += int(is_auto_version_query)

                return data

            insert_race_condition.count = 0

            monkeypatch.setattr(
                get_storage().__class__, "fetch_experiments", insert_race_condition
            )

            experiment = create_experiment(
                config["name"], space={"y": "uniform(0, 10)"}
            )

            assert insert_race_condition.count == 1
            assert experiment.version == 2
            assert experiment.configuration == child.configuration

    def test_create_experiment_race_condition_broken(self, monkeypatch):
        """Test that two or more race condition leads to raise"""
        with OrionState(experiments=[config]):
            parent = create_experiment(config["name"])
            child = create_experiment(config["name"], space={"y": "uniform(0, 10)"})

            def insert_race_condition(self, query):
                is_auto_version_query = query == {
                    "name": config["name"],
                    "refers.parent_id": parent.id,
                }
                if is_auto_version_query:
                    data = [child.configuration]
                # The query returns no other child, never!
                else:
                    data = [parent.configuration]

                insert_race_condition.count += int(is_auto_version_query)

                return data

            insert_race_condition.count = 0

            monkeypatch.setattr(
                get_storage().__class__, "fetch_experiments", insert_race_condition
            )

            with pytest.raises(RaceCondition) as exc:
                create_experiment(config["name"], space={"y": "uniform(0, 10)"})

            assert insert_race_condition.count == 2
            assert "There was a race condition during branching and new version" in str(
                exc.value
            )

    def test_create_experiment_hit_manual_branch(self):
        """Test creating a differing experiment that cause branching."""
        new_space = {"y": "uniform(0, 10)"}
        with OrionState(experiments=[config]):
            create_experiment(config["name"], space=new_space)

            with pytest.raises(BranchingEvent) as exc:
                create_experiment(config["name"], version=1, space=new_space)

            assert "Configuration is different and generates" in str(exc.value)

    def test_create_experiment_debug_mode(self, tmp_path):
        """Test that EphemeralDB is used in debug mode whatever the storage config given"""
        update_singletons()

        conf_file = str(tmp_path / "db.pkl")

        create_experiment(
            config["name"],
            space={"x": "uniform(0, 10)"},
            storage={
                "type": "legacy",
                "database": {"type": "pickleddb", "host": conf_file},
            },
        )

        storage = get_storage()

        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)

        update_singletons()

        create_experiment(
            config["name"],
            space={"x": "uniform(0, 10)"},
            storage={"type": "legacy", "database": {"type": "pickleddb"}},
            debug=True,
        )

        storage = get_storage()

        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, EphemeralDB)


class TestWorkon:
    """Test the helper function for sequential API"""

    def test_workon(self):
        """Verify that workon processes properly"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        experiment = workon(foo, space={"x": "uniform(0, 10)"}, max_trials=5)
        assert len(experiment.fetch_trials()) == 5
        assert experiment.name == "loop"
        assert isinstance(experiment._experiment._storage, Legacy)
        assert isinstance(experiment._experiment._storage._db, EphemeralDB)

    def test_workon_algo(self):
        """Verify that algo config is processed properly"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        experiment = workon(
            foo,
            space={"x": "uniform(0, 10)"},
            max_trials=5,
            algorithms={"random": {"seed": 5}},
        )

        assert experiment.algorithms.algorithm.seed == 5

    def test_workon_name(self):
        """Verify setting the name with workon"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        experiment = workon(
            foo, space={"x": "uniform(0, 10)"}, max_trials=5, name="voici"
        )

        assert experiment.name == "voici"

    def test_workon_fail(self, monkeypatch):
        """Verify that storage is reverted if workon fails"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        def build_fail(*args, **kwargs):
            raise RuntimeError("You shall not build!")

        monkeypatch.setattr("orion.core.io.experiment_builder.build", build_fail)

        # Flush storage singleton
        update_singletons()

        with pytest.raises(RuntimeError) as exc:
            experiment = workon(
                foo, space={"x": "uniform(0, 10)"}, max_trials=5, name="voici"
            )

        assert exc.match("You shall not build!")

        # Verify that tmp storage was cleared
        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        # Now test with a prior storage
        with OrionState(
            storage={"type": "legacy", "database": {"type": "EphemeralDB"}}
        ):
            storage = get_storage()

            with pytest.raises(RuntimeError) as exc:
                workon(foo, space={"x": "uniform(0, 10)"}, max_trials=5, name="voici")

            assert exc.match("You shall not build!")

            assert get_storage() is storage

    def test_workon_twice(self):
        """Verify setting the each experiment has its own storage"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        experiment = workon(
            foo, space={"x": "uniform(0, 10)"}, max_trials=5, name="voici"
        )

        assert experiment.name == "voici"
        assert len(experiment.fetch_trials()) == 5

        experiment2 = workon(
            foo, space={"x": "uniform(0, 10)"}, max_trials=1, name="voici"
        )

        assert experiment2.name == "voici"
        assert len(experiment2.fetch_trials()) == 1

    def test_workon_with_parallel_backend(self):
        """Test there is no impact of joblib parallel for workon function"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        import joblib

        with joblib.parallel_backend("loky"):
            experiment = workon(
                foo, space={"x": "uniform(0, 10)"}, max_trials=5, name="voici"
            )

        assert experiment.name == "voici"
        assert len(experiment.fetch_trials()) == 5

        with joblib.parallel_backend("loky", n_jobs=-1):
            experiment = workon(
                foo, space={"x": "uniform(0, 10)"}, max_trials=3, name="voici"
            )

        assert experiment.name == "voici"
        assert len(experiment.fetch_trials()) == 3


class TestGetExperiment:
    """Test :meth:`orion.client.get_experiment`"""

    @pytest.mark.usefixtures("mock_database")
    def test_experiment_do_not_exist(self):
        """Tests that an error is returned when the experiment doesn't exist"""
        with pytest.raises(NoConfigurationError) as exception:
            get_experiment("a")
        assert (
            "No experiment with given name 'a' and version '*' inside database, "
            "no view can be created." == str(exception.value)
        )

    @pytest.mark.usefixtures("mock_database")
    def test_experiment_exist(self):
        """
        Tests that an instance of :class:`orion.client.experiment.ExperimentClient` is
        returned representing the latest version when none is given.
        """
        experiment = create_experiment("a", space={"x": "uniform(0, 10)"})

        experiment = get_experiment("a")

        assert experiment
        assert isinstance(experiment, ExperimentClient)
        assert experiment.mode == "r"

    @pytest.mark.usefixtures("mock_database")
    def test_version_do_not_exist(self, caplog):
        """Tests that a warning is printed when the experiment exist but the version doesn't"""
        create_experiment("a", space={"x": "uniform(0, 10)"})

        experiment = get_experiment("a", 2)

        assert experiment.version == 1
        assert (
            "Version 2 was specified but most recent version is only 1. Using 1."
            in caplog.text
        )

    @pytest.mark.usefixtures("mock_database")
    def test_read_write_mode(self):
        """Tests that experiment can be created in write mode"""
        experiment = create_experiment("a", space={"x": "uniform(0, 10)"})
        assert experiment.mode == "x"

        experiment = get_experiment("a", 2, mode="r")
        assert experiment.mode == "r"

        with pytest.raises(UnsupportedOperation) as exc:
            experiment.insert({"x": 0})

        assert exc.match("ExperimentClient must have write rights to execute `insert()")

        experiment = get_experiment("a", 2, mode="w")
        assert experiment.mode == "w"

        trial = experiment.insert({"x": 0})

        with pytest.raises(UnsupportedOperation) as exc:
            experiment.reserve(trial)

        assert exc.match(
            "ExperimentClient must have execution rights to execute `reserve()"
        )



"""Common fixtures and utils for configuration tests."""
from orion.algo.base import BaseAlgorithm, OptimizationAlgorithm
from orion.core.worker.strategy import BaseParallelStrategy, Strategy


def __init__(self, *args, **params):
    """Initialize the parent class"""
    self.params = params
    super(type(self), self).__init__(*args, **params)


def stub(self, *args, **kwargs):
    """Just a stub..."""
    pass


def configuration(self):
    """Configuration based on self.params"""
    return {self.__class__.__name__.lower(): self.params}


for char in "ABCDE":
    algo_class = type(f"A{char}", (BaseAlgorithm,), {"suggest": stub, "observe": stub})

    # Hack it into being discoverable
    OptimizationAlgorithm.types[algo_class.__name__.lower()] = algo_class


for char in "ABCDE":
    strategy_class = type(
        f"S{char}", (BaseParallelStrategy,), {"observe": stub, "__init__": __init__}
    )

    strategy_class.configuration = property(configuration)

    # Hack it into being discoverable
    Strategy.types[strategy_class.__name__.lower()] = strategy_class



"""Example usage and tests for :mod:`orion.core.utils.backward`."""
import pytest

import orion.core.utils.backward as backward
from orion.algo.base import BaseAlgorithm


def create_algo_class(**attributes):
    """Create algo class with given requirements attributes"""
    return type("Algo", (BaseAlgorithm,), attributes)


TYPE_REQUIREMENTS = ["real", "integer", "numerical"]


@pytest.mark.parametrize("requirement", TYPE_REQUIREMENTS)
def test_type_requirements(requirement):
    """Test algorithms type requirement defined in old requirements API"""
    algo_class = create_algo_class(requires=[requirement])
    requirements = backward.get_algo_requirements(algo_class)
    assert len(requirements) == 3
    assert requirements == {
        "type_requirement": requirement,
        "shape_requirement": None,
        "dist_requirement": None,
    }


@pytest.mark.parametrize("requirement", TYPE_REQUIREMENTS)
def test_shape_requirements(requirement):
    """Test algorithms shape requirement defined in old requirements API"""
    algo_class = create_algo_class(requires=[requirement, "flattened"])
    requirements = backward.get_algo_requirements(algo_class)
    assert len(requirements) == 3
    assert requirements == {
        "type_requirement": requirement,
        "shape_requirement": "flattened",
        "dist_requirement": None,
    }


@pytest.mark.parametrize("requirement", TYPE_REQUIREMENTS)
def test_dist_requirements(requirement):
    """Test algorithms dist requirement defined in old requirements API"""
    algo_class = create_algo_class(requires=[requirement, "linear"])
    requirements = backward.get_algo_requirements(algo_class)
    assert len(requirements) == 3
    assert requirements == {
        "type_requirement": requirement,
        "shape_requirement": None,
        "dist_requirement": "linear",
    }


@pytest.mark.parametrize("requirement", TYPE_REQUIREMENTS)
def test_no_changes(requirement):
    """Test algorithms following new requirements API"""
    algo_class = create_algo_class(
        requires_type=requirement, requires_shape="flattened", requires_dist="linear"
    )
    requirements = backward.get_algo_requirements(algo_class)
    assert len(requirements) == 3
    assert requirements == {
        "type_requirement": requirement,
        "shape_requirement": "flattened",
        "dist_requirement": "linear",
    }
